import datetime
import json
from pathlib import Path
from typing import Optional

from dateutil.parser import parse as dateparse
from pydantic import DirectoryPath, FilePath, validate_call

from ..baseimagingextractorinterface import BaseImagingExtractorInterface


class ScanImageImagingInterface(BaseImagingExtractorInterface):
    """
    Interface for reading TIFF files produced via ScanImage.

    It extracts metadata from the provided TIFF file and determines the ScanImage version.
    For the legacy version 3.8, it creates an instance of ScanImageLegacyImagingInterface.
    For newer versions, it parses the metadata and determines the number of planes.
    If there is more than one plane and no specific plane is provided, it creates an instance of ScanImageMultiPlaneImagingInterface.
    If there is only one plane or a specific plane is provided, it creates an instance of ScanImageSinglePlaneImagingInterface.
    """

    display_name = "ScanImage Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage TIFF files."

    ExtractorName = "ScanImageTiffImagingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to Tiff file."
        return source_schema

    @validate_call
    def __new__(
        cls,
        file_path: FilePath,
        channel_name: Optional[str] = None,
        plane_name: Optional[str] = None,
        fallback_sampling_frequency: Optional[float] = None,
        verbose: bool = True,
    ):
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        image_metadata = extract_extra_metadata(file_path=file_path)
        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            return ScanImageLegacyImagingInterface(
                file_path=file_path,
                fallback_sampling_frequency=fallback_sampling_frequency,
                verbose=verbose,
            )

        parsed_metadata = parse_metadata(metadata=image_metadata)
        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if len(available_planes) > 1 and plane_name is None:
            return ScanImageMultiPlaneImagingInterface(
                file_path=file_path,
                channel_name=channel_name,
                image_metadata=image_metadata,
                parsed_metadata=parsed_metadata,
                verbose=verbose,
            )

        return ScanImageSinglePlaneImagingInterface(
            file_path=file_path,
            channel_name=channel_name,
            plane_name=plane_name,
            image_metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            verbose=verbose,
        )


class ScanImageLegacyImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading TIFF files produced via ScanImage v3.8."""

    display_name = "ScanImage Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage v3.8 TIFF files."

    ExtractorName = "ScanImageTiffImagingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to Tiff file."
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        fallback_sampling_frequency: Optional[float] = None,
        verbose: bool = True,
    ):
        """
        DataInterface for reading Tiff files that are generated by ScanImage v3.8. This interface extracts the metadata
        from the exif of the tiff file.

        Parameters
        ----------
        file_path: FilePath
            Path to tiff file.
        fallback_sampling_frequency: float, optional
            The sampling frequency can usually be extracted from the scanimage metadata in
            exif:ImageDescription:state.acq.frameRate. If not, use this.
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
        )

        self.image_metadata = extract_extra_metadata(file_path=file_path)

        if "state.acq.frameRate" in self.image_metadata:
            sampling_frequency = float(self.image_metadata["state.acq.frameRate"])
        elif "SI.hRoiManager.scanFrameRate" in self.image_metadata:
            sampling_frequency = float(self.image_metadata["SI.hRoiManager.scanFrameRate"])
        else:
            assert_msg = (
                "sampling frequency not found in image metadata, "
                "input the frequency using the argument `fallback_sampling_frequency`"
            )
            assert fallback_sampling_frequency is not None, assert_msg
            sampling_frequency = fallback_sampling_frequency

        super().__init__(file_path=file_path, sampling_frequency=sampling_frequency, verbose=verbose)

    def get_metadata(self) -> dict:
        device_number = 0  # Imaging plane metadata is a list with metadata for each plane

        metadata = super().get_metadata()

        if "state.internal.triggerTimeString" in self.image_metadata:
            extracted_session_start_time = dateparse(self.image_metadata["state.internal.triggerTimeString"])
            metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        # Extract many scan image properties and attach them as dic in the description
        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][device_number]
        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        return metadata


class ScanImageMultiFileImagingInterface(BaseImagingExtractorInterface):
    """
    Interface for reading multi-file (buffered) TIFF files produced via ScanImage.

    It extracts metadata from the first TIFF file in the provided folder and determines the number of available planes.
    If there is more than one plane and no specific plane is provided, it creates an instance of ScanImageMultiPlaneMultiFileImagingInterface.
    If there is only one plane or a specific plane is provided, it creates an instance of ScanImageSinglePlaneMultiFileImagingInterface.
    """

    display_name = "ScanImage Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneMultiFileImagingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"]["description"] = "Path to the folder containing the TIFF files."
        return source_schema

    @validate_call
    def __new__(
        cls,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: Optional[str] = None,
        plane_name: Optional[str] = None,
        extract_all_metadata: bool = False,
        verbose: bool = True,
    ):
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = extract_extra_metadata(file_path=first_file_path)

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError("ScanImage version 3.8 is not supported.")

        parsed_metadata = parse_metadata(metadata=image_metadata)
        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if len(available_planes) > 1 and plane_name is None:
            return ScanImageMultiPlaneMultiFileImagingInterface(
                folder_path=folder_path,
                file_pattern=file_pattern,
                channel_name=channel_name,
                image_metadata=image_metadata,
                parsed_metadata=parsed_metadata,
                extract_all_metadata=extract_all_metadata,
                verbose=verbose,
            )

        return ScanImageSinglePlaneMultiFileImagingInterface(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            plane_name=plane_name,
            image_metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )


class ScanImageMultiPlaneImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading multi plane (volumetric) TIFF files produced via ScanImage."""

    display_name = "ScanImage Volumetric Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi plane (volumetric) TIFF files."

    ExtractorName = "ScanImageTiffMultiPlaneImagingExtractor"

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        channel_name: Optional[str] = None,
        image_metadata: Optional[dict] = None,
        parsed_metadata: Optional[dict] = None,
        verbose: bool = True,
    ):
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        file_path : FilePath
            Path to the TIFF file.
        channel_name : str
            Name of the channel for this extractor.
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        image_metadata = image_metadata or extract_extra_metadata(file_path=file_path)
        self.image_metadata = image_metadata
        parsed_metadata = parsed_metadata or parse_metadata(metadata=self.image_metadata)

        if parsed_metadata["num_planes"] == 1:
            raise ValueError(
                "Only one plane detected. For single plane imaging data use ScanImageSinglePlaneImagingInterface instead."
            )

        avaliable_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(avaliable_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! Please specify which channel you wish to load "
                    "with the `channel_name` argument. To see which channels are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...)`"
                )
            channel_name = avaliable_channels[0]
        assert channel_name in avaliable_channels, f"Channel '{channel_name}' not found in the tiff file."

        two_photon_series_name_suffix = None
        if len(avaliable_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            file_path=file_path,
            channel_name=channel_name,
            metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageMultiPlaneMultiFileImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading volumetric multi-file (buffered) TIFF files produced via ScanImage."""

    display_name = "ScanImage Volumetric Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) volumetric TIFF files."

    ExtractorName = "ScanImageTiffMultiPlaneMultiFileImagingExtractor"

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: Optional[str] = None,
        extract_all_metadata: bool = False,
        image_metadata: Optional[dict] = None,
        parsed_metadata: Optional[dict] = None,
        verbose: bool = True,
    ):
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing the TIFF files.
        file_pattern : str
            Pattern for the TIFF files to read -- see pathlib.Path.glob for details.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        extract_all_metadata : bool
            If True, extract metadata from every file in the folder. If False, only extract metadata from the first
            file in the folder. The default is False.
        """
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = image_metadata or extract_extra_metadata(file_path=first_file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError("ScanImage version 3.8 is not supported. Please use ScanImageImagingInterface instead.")

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        if parsed_metadata["num_planes"] == 1:
            raise ValueError(
                "Only one plane detected. For single plane imaging data use ScanImageSinglePlaneMultiFileImagingInterface instead."
            )

        avaliable_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(avaliable_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! Please specify which channel you wish to load "
                    "with the `channel_name` argument. To see which channels are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...)`"
                )
            channel_name = avaliable_channels[0]
        assert channel_name in avaliable_channels, f"Channel '{channel_name}' not found in the tiff file."

        two_photon_series_name_suffix = None
        if len(avaliable_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageSinglePlaneImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading TIFF files produced via ScanImage."""

    display_name = "ScanImage Single Plane Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneImagingExtractor"

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        channel_name: Optional[str] = None,
        plane_name: Optional[str] = None,
        image_metadata: Optional[dict] = None,
        parsed_metadata: Optional[dict] = None,
        verbose: bool = True,
    ):
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        file_path : FilePath
            Path to the TIFF file.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        plane_name : str
            The name of the plane to load, to determine what planes are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...).
        """
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        image_metadata = image_metadata or extract_extra_metadata(file_path=file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError("ScanImage version 3.8 is not supported. Please use ScanImageImagingInterface instead.")

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        avaliable_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(avaliable_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! Please specify which channel you wish to load "
                    "with the `channel_name` argument. To see which channels are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...)`"
                )
            channel_name = avaliable_channels[0]
        assert channel_name in avaliable_channels, f"Channel '{channel_name}' not found in the tiff file."

        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if plane_name is None:
            if len(available_planes) > 1:
                raise ValueError(
                    "More than one plane is detected! Please specify which plane you wish to load "
                    "with the `plane_name` argument. To see which planes are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...)`"
                )
            plane_name = available_planes[0]
        assert plane_name in available_planes, f"Plane '{plane_name}' not found in the tiff file."

        two_photon_series_name_suffix = None
        if len(avaliable_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        if len(available_planes) > 1:
            two_photon_series_name_suffix = f"{two_photon_series_name_suffix}Plane{plane_name}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            file_path=file_path,
            channel_name=channel_name,
            plane_name=plane_name,
            metadata=image_metadata,
            parsed_metadata=parsed_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


class ScanImageSinglePlaneMultiFileImagingInterface(BaseImagingExtractorInterface):
    """Interface for reading multi-file (buffered) TIFF files produced via ScanImage."""

    display_name = "ScanImage Single Plane Multi-File Imaging"
    associated_suffixes = (".tif",)
    info = "Interface for ScanImage multi-file (buffered) TIFF files."

    ExtractorName = "ScanImageTiffSinglePlaneMultiFileImagingExtractor"

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        file_pattern: str,
        channel_name: Optional[str] = None,
        plane_name: Optional[str] = None,
        image_metadata: Optional[dict] = None,
        parsed_metadata: Optional[dict] = None,
        extract_all_metadata: bool = False,
        verbose: bool = True,
    ):
        """
        DataInterface for reading multi-file (buffered) TIFF files that are generated by ScanImage.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing the TIFF files.
        file_pattern : str
            Pattern for the TIFF files to read -- see pathlib.Path.glob for details.
        channel_name : str
            The name of the channel to load, to determine what channels are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...).
        plane_name : str
            The name of the plane to load, to determine what planes are available use ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...).
        extract_all_metadata : bool
            If True, extract metadata from every file in the folder. If False, only extract metadata from the first
            file in the folder. The default is False.
        """
        from natsort import natsorted
        from roiextractors.extractors.tiffimagingextractors.scanimagetiff_utils import (
            extract_extra_metadata,
            parse_metadata,
        )

        file_paths = natsorted(Path(folder_path).glob(file_pattern))
        first_file_path = file_paths[0]

        image_metadata = image_metadata or extract_extra_metadata(file_path=first_file_path)
        self.image_metadata = image_metadata

        version = get_scanimage_major_version(scanimage_metadata=image_metadata)
        if version == "3.8":
            raise ValueError("ScanImage version 3.8 is not supported. Please use ScanImageImagingInterface instead.")

        parsed_metadata = parsed_metadata or parse_metadata(metadata=image_metadata)
        avaliable_channels = parsed_metadata["channel_names"]
        if channel_name is None:
            if len(avaliable_channels) > 1:
                raise ValueError(
                    "More than one channel is detected! Please specify which channel you wish to load "
                    "with the `channel_name` argument. To see which channels are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_channels(file_path=...)`"
                )
            channel_name = avaliable_channels[0]
        assert channel_name in avaliable_channels, f"Channel '{channel_name}' not found in the tiff file."

        available_planes = [f"{i}" for i in range(parsed_metadata["num_planes"])]
        if plane_name is None:
            if len(available_planes) > 1:
                raise ValueError(
                    "More than one plane is detected! Please specify which plane you wish to load "
                    "with the `plane_name` argument. To see which planes are available, use "
                    "`ScanImageTiffSinglePlaneImagingExtractor.get_available_planes(file_path=...)`"
                )
            plane_name = available_planes[0]
        assert plane_name in available_planes, f"Plane '{plane_name}' not found in the tiff file."

        two_photon_series_name_suffix = None
        if len(avaliable_channels) > 1:
            two_photon_series_name_suffix = f"{channel_name.replace(' ', '')}"
        if len(available_planes) > 1:
            two_photon_series_name_suffix = f"{two_photon_series_name_suffix}Plane{plane_name}"
        self.two_photon_series_name_suffix = two_photon_series_name_suffix

        super().__init__(
            folder_path=folder_path,
            file_pattern=file_pattern,
            channel_name=channel_name,
            plane_name=plane_name,
            extract_all_metadata=extract_all_metadata,
            verbose=verbose,
        )

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        extracted_session_start_time = datetime.datetime.strptime(
            self.image_metadata["epoch"], "[%Y %m %d %H %M %S.%f]"
        )
        metadata["NWBFile"].update(session_start_time=extracted_session_start_time)

        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][0]

        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        if self.two_photon_series_name_suffix is None:
            return metadata

        imaging_plane_metadata = ophys_metadata["ImagingPlane"][0]
        channel_name = self.source_data["channel_name"]
        optical_channel_metadata = [
            channel for channel in imaging_plane_metadata["optical_channel"] if channel["name"] == channel_name
        ]
        imaging_plane_name = f"ImagingPlane{self.two_photon_series_name_suffix}"
        imaging_plane_metadata.update(
            name=imaging_plane_name,
            optical_channel=optical_channel_metadata,
        )
        two_photon_series_metadata.update(
            name=f"TwoPhotonSeries{self.two_photon_series_name_suffix}",
            imaging_plane=imaging_plane_name,
        )

        return metadata


def get_scanimage_major_version(scanimage_metadata: dict) -> str:
    """
    Determine the version of ScanImage that produced the TIFF file.

    Parameters
    ----------
    scanimage_metadata : dict
        Dictionary of metadata extracted from a TIFF file produced via ScanImage.

    Returns
    -------
    version : str
        The version of ScanImage that produced the TIFF file.
    """
    if "SI.VERSION_MAJOR" in scanimage_metadata:
        return scanimage_metadata["SI.VERSION_MAJOR"]
    elif "state.software.version" in scanimage_metadata:
        return scanimage_metadata["state.software.version"]

    raise ValueError("ScanImage version could not be determined from metadata.")



from pydantic import FilePath

from ..basesegmentationextractorinterface import BaseSegmentationExtractorInterface


class CaimanSegmentationInterface(BaseSegmentationExtractorInterface):
    """Data interface for CaimanSegmentationExtractor."""

    display_name = "CaImAn Segmentation"
    associated_suffixes = (".hdf5",)
    info = "Interface for Caiman segmentation data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_metadata = super().get_source_schema()
        source_metadata["properties"]["file_path"]["description"] = "Path to .hdf5 file."
        return source_metadata

    def __init__(self, file_path: FilePath, verbose: bool = True):
        """

        Parameters
        ----------
        file_path : FilePath
            Path to .hdf5 file.
        verbose : bool, default True
            Whether to print progress
        """
        super().__init__(file_path=file_path)
        self.verbose = verbose



# Behavior
from .behavior.audio.audiointerface import AudioInterface
from .behavior.deeplabcut.deeplabcutdatainterface import DeepLabCutInterface
from .behavior.fictrac.fictracdatainterface import FicTracDataInterface
from .behavior.lightningpose.lightningposedatainterface import (
    LightningPoseDataInterface,
)
from .behavior.medpc.medpcdatainterface import MedPCInterface
from .behavior.miniscope.miniscopedatainterface import MiniscopeBehaviorInterface
from .behavior.neuralynx.neuralynx_nvt_interface import NeuralynxNvtInterface
from .behavior.sleap.sleapdatainterface import SLEAPInterface
from .behavior.video.videodatainterface import VideoInterface

# Ecephys
from .ecephys.alphaomega.alphaomegadatainterface import AlphaOmegaRecordingInterface
from .ecephys.axona.axonadatainterface import (
    AxonaLFPDataInterface,
    AxonaPositionDataInterface,
    AxonaRecordingInterface,
    AxonaUnitRecordingInterface,
)
from .ecephys.biocam.biocamdatainterface import BiocamRecordingInterface
from .ecephys.blackrock.blackrockdatainterface import (
    BlackrockRecordingInterface,
    BlackrockSortingInterface,
)
from .ecephys.cellexplorer.cellexplorerdatainterface import (
    CellExplorerLFPInterface,
    CellExplorerRecordingInterface,
    CellExplorerSortingInterface,
)
from .ecephys.edf.edfdatainterface import EDFRecordingInterface
from .ecephys.intan.intandatainterface import IntanRecordingInterface
from .ecephys.kilosort.kilosortdatainterface import KiloSortSortingInterface
from .ecephys.maxwell.maxonedatainterface import MaxOneRecordingInterface
from .ecephys.mcsraw.mcsrawdatainterface import MCSRawRecordingInterface
from .ecephys.mearec.mearecdatainterface import MEArecRecordingInterface
from .ecephys.neuralynx.neuralynxdatainterface import (
    NeuralynxRecordingInterface,
    NeuralynxSortingInterface,
)
from .ecephys.neuroscope.neuroscopedatainterface import (
    NeuroScopeLFPInterface,
    NeuroScopeRecordingInterface,
    NeuroScopeSortingInterface,
)
from .ecephys.openephys.openephysbinarydatainterface import (
    OpenEphysBinaryRecordingInterface,
)
from .ecephys.openephys.openephysdatainterface import OpenEphysRecordingInterface
from .ecephys.openephys.openephyslegacydatainterface import (
    OpenEphysLegacyRecordingInterface,
)
from .ecephys.openephys.openephyssortingdatainterface import OpenEphysSortingInterface
from .ecephys.phy.phydatainterface import PhySortingInterface
from .ecephys.plexon.plexondatainterface import (
    Plexon2RecordingInterface,
    PlexonRecordingInterface,
    PlexonSortingInterface,
)
from .ecephys.spike2.spike2datainterface import Spike2RecordingInterface
from .ecephys.spikegadgets.spikegadgetsdatainterface import (
    SpikeGadgetsRecordingInterface,
)
from .ecephys.spikeglx.spikeglxdatainterface import SpikeGLXRecordingInterface
from .ecephys.spikeglx.spikeglxnidqinterface import SpikeGLXNIDQInterface
from .ecephys.tdt.tdtdatainterface import TdtRecordingInterface

# Icephys
from .icephys.abf.abfdatainterface import AbfInterface

# Ophys
from .ophys.brukertiff.brukertiffdatainterface import (
    BrukerTiffMultiPlaneImagingInterface,
    BrukerTiffSinglePlaneImagingInterface,
)
from .ophys.caiman.caimandatainterface import CaimanSegmentationInterface
from .ophys.cnmfe.cnmfedatainterface import CnmfeSegmentationInterface
from .ophys.extract.extractdatainterface import ExtractSegmentationInterface
from .ophys.hdf5.hdf5datainterface import Hdf5ImagingInterface
from .ophys.micromanagertiff.micromanagertiffdatainterface import (
    MicroManagerTiffImagingInterface,
)
from .ophys.miniscope.miniscopeimagingdatainterface import MiniscopeImagingInterface
from .ophys.sbx.sbxdatainterface import SbxImagingInterface
from .ophys.scanimage.scanimageimaginginterfaces import (
    ScanImageImagingInterface,
    ScanImageMultiFileImagingInterface,
)
from .ophys.sima.simadatainterface import SimaSegmentationInterface
from .ophys.suite2p.suite2pdatainterface import Suite2pSegmentationInterface
from .ophys.tdt_fp.tdtfiberphotometrydatainterface import TDTFiberPhotometryInterface
from .ophys.tiff.tiffdatainterface import TiffImagingInterface

# Text
from .text.csv.csvtimeintervalsinterface import CsvTimeIntervalsInterface
from .text.excel.exceltimeintervalsinterface import ExcelTimeIntervalsInterface

interface_list = [
    # Ecephys
    NeuralynxRecordingInterface,
    NeuralynxSortingInterface,
    NeuroScopeRecordingInterface,
    NeuroScopeSortingInterface,
    NeuroScopeLFPInterface,
    Spike2RecordingInterface,
    SpikeGLXRecordingInterface,
    SpikeGLXNIDQInterface,
    SpikeGadgetsRecordingInterface,
    IntanRecordingInterface,
    CellExplorerSortingInterface,
    CellExplorerRecordingInterface,
    CellExplorerLFPInterface,
    BlackrockRecordingInterface,
    BlackrockSortingInterface,
    OpenEphysRecordingInterface,
    OpenEphysBinaryRecordingInterface,
    OpenEphysLegacyRecordingInterface,
    OpenEphysSortingInterface,
    PhySortingInterface,
    KiloSortSortingInterface,
    AxonaRecordingInterface,
    AxonaPositionDataInterface,
    AxonaLFPDataInterface,
    AxonaUnitRecordingInterface,
    EDFRecordingInterface,
    TdtRecordingInterface,
    PlexonRecordingInterface,
    Plexon2RecordingInterface,
    PlexonSortingInterface,
    BiocamRecordingInterface,
    AlphaOmegaRecordingInterface,
    MEArecRecordingInterface,
    MCSRawRecordingInterface,
    MaxOneRecordingInterface,
    # Icephys
    AbfInterface,
    # Ophys
    CaimanSegmentationInterface,
    CnmfeSegmentationInterface,
    Suite2pSegmentationInterface,
    ExtractSegmentationInterface,
    SimaSegmentationInterface,
    SbxImagingInterface,
    TiffImagingInterface,
    Hdf5ImagingInterface,
    ScanImageImagingInterface,
    ScanImageMultiFileImagingInterface,
    BrukerTiffMultiPlaneImagingInterface,
    BrukerTiffSinglePlaneImagingInterface,
    MicroManagerTiffImagingInterface,
    MiniscopeImagingInterface,
    TDTFiberPhotometryInterface,
    # Behavior
    VideoInterface,
    AudioInterface,
    DeepLabCutInterface,
    SLEAPInterface,
    MiniscopeBehaviorInterface,
    FicTracDataInterface,
    NeuralynxNvtInterface,
    LightningPoseDataInterface,
    MedPCInterface,
    # Text
    CsvTimeIntervalsInterface,
    ExcelTimeIntervalsInterface,
]

interfaces_by_category = dict(
    ecephys={
        interface.__name__.replace("RecordingInterface", ""): interface  # TODO: use removesuffix when 3.8 is dropped
        for interface in interface_list
        if "Recording" in interface.__name__
    },
    sorting={
        interface.__name__.replace("SortingInterface", ""): interface
        for interface in interface_list
        if "Sorting" in interface.__name__
    },
    imaging={
        interface.__name__.replace("ImagingInterface", ""): interface
        for interface in interface_list
        if "Imaging" in interface.__name__
    },
    segmentation={
        interface.__name__.replace("SegmentationInterface", ""): interface
        for interface in interface_list
        if "Segmentation" in interface.__name__
    },
    fiber_photometry={"TDTFiberPhotometry": TDTFiberPhotometryInterface},
    icephys=dict(Abf=AbfInterface),
    behavior=dict(
        Video=VideoInterface,
        DeepLabCut=DeepLabCutInterface,
        SLEAP=SLEAPInterface,
        FicTrac=FicTracDataInterface,
        LightningPose=LightningPoseDataInterface,
        # Text
        CsvTimeIntervals=CsvTimeIntervalsInterface,
        ExcelTimeIntervals=ExcelTimeIntervalsInterface,
        MedPC=MedPCInterface,
    ),
)



from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from pydantic_core import PydanticOmit
from pydantic_core.core_schema import CoreSchema


class PureJSONSchemaGenerator(GenerateJsonSchema):
    """
    Sometimes our Pydantic models include valid Pythonic types, such as `IsInstance[SomeGenericClass]`.

    However, these cases have no valid JSON equivalent as they are purely for programmatic usage.
    Thus, we exclude them from the schema generation.
    """

    def handle_invalid_for_json_schema(self, schema: CoreSchema, error_info: str) -> JsonSchemaValue:
        raise PydanticOmit



"""Authors: Heberto Mayorquin, Cody Baker and Ben Dichter."""

from typing import Optional

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ophys import Fluorescence, ImageSegmentation, ImagingPlane, TwoPhotonSeries

from ...baseextractorinterface import BaseExtractorInterface
from ...utils import fill_defaults, get_base_schema, get_schema_from_hdmf_class


class BaseSegmentationExtractorInterface(BaseExtractorInterface):
    """Parent class for all SegmentationExtractorInterfaces."""

    keywords = ("segmentation", "roi", "cells")

    ExtractorModuleName = "roiextractors"

    def __init__(self, verbose: bool = False, **source_data):
        super().__init__(**source_data)
        self.verbose = verbose
        self.segmentation_extractor = self.get_extractor()(**source_data)

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()
        metadata_schema["required"] = ["Ophys"]
        metadata_schema["properties"]["Ophys"] = get_base_schema()
        metadata_schema["properties"]["Ophys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items=get_schema_from_hdmf_class(Device)),
        )
        metadata_schema["properties"]["Ophys"]["properties"].update(
            Fluorescence=get_schema_from_hdmf_class(Fluorescence),
            ImageSegmentation=get_schema_from_hdmf_class(ImageSegmentation),
            ImagingPlane=get_schema_from_hdmf_class(ImagingPlane),
            TwoPhotonSeries=get_schema_from_hdmf_class(TwoPhotonSeries),
        )
        metadata_schema["properties"]["Ophys"]["required"] = ["Device", "ImageSegmentation"]

        # Temporary fixes until centralized definition of metadata schemas
        metadata_schema["properties"]["Ophys"]["properties"]["ImagingPlane"].update(type="array")
        metadata_schema["properties"]["Ophys"]["properties"]["TwoPhotonSeries"].update(type="array")

        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].update(required=["name"])
        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].pop("additionalProperties")

        roi_response_series_schema = metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"][
            "properties"
        ].pop("roi_response_series")

        roi_response_series_schema.pop("maxItems")
        roi_response_series_schema["items"].update(required=list())

        roi_response_series_per_plane_schema = dict(
            type="object", patternProperties={"^[a-zA-Z0-9]+$": roi_response_series_schema["items"]}
        )

        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"].update(
            patternProperties={"^(?!name$)[a-zA-Z0-9]+$": roi_response_series_per_plane_schema}
        )

        metadata_schema["properties"]["Ophys"]["properties"]["ImageSegmentation"]["additionalProperties"] = True

        metadata_schema["properties"]["Ophys"]["properties"]["DfOverF"] = metadata_schema["properties"]["Ophys"][
            "properties"
        ]["Fluorescence"]

        # NOTE: Would prefer to remove in favor of simply using the up-to-date metadata_schema.json
        images_inner_schema = dict(
            type="object",
            properties=dict(name=dict(type="string"), description=dict(type="string")),
        )

        summary_images_per_plane_schema = dict(type="object", patternProperties={"^[a-zA-Z0-9]+$": images_inner_schema})

        metadata_schema["properties"]["Ophys"]["properties"]["SegmentationImages"] = dict(
            type="object",
            required=["name"],
            properties=dict(
                name=dict(type="string", default="SegmentationImages"),
                description=dict(type="string"),
            ),
            patternProperties={
                "^(?!(name|description)$)[a-zA-Z0-9]+$": summary_images_per_plane_schema,
            },
        )

        fill_defaults(metadata_schema, self.get_metadata())
        return metadata_schema

    def get_metadata(self) -> dict:
        from ...tools.roiextractors import get_nwb_segmentation_metadata

        metadata = super().get_metadata()
        metadata.update(get_nwb_segmentation_metadata(self.segmentation_extractor))
        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        reinitialized_extractor = self.get_extractor()(**self.source_data)
        return reinitialized_extractor.frame_to_time(frames=np.arange(stop=reinitialized_extractor.get_num_frames()))

    def get_timestamps(self) -> np.ndarray:
        return self.segmentation_extractor.frame_to_time(
            frames=np.arange(stop=self.segmentation_extractor.get_num_frames())
        )

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        self.segmentation_extractor.set_times(times=aligned_timestamps)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        stub_test: bool = False,
        stub_frames: int = 100,
        include_background_segmentation: bool = False,
        include_roi_centroids: bool = True,
        include_roi_acceptance: bool = True,
        mask_type: Optional[str] = "image",  # Literal["image", "pixel", "voxel"]
        plane_segmentation_name: Optional[str] = None,
        iterator_options: Optional[dict] = None,
        compression_options: Optional[
            dict
        ] = None,  # TODO: remove completely after 10/1/2024; still passing for deprecation warning
    ):
        """

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile to add the plane segmentation to.
        metadata : dict, optional
            The metadata for the interface
        stub_test : bool, default: False
        stub_frames : int, default: 100
        include_background_segmentation : bool, default: False
            Whether to include the background plane segmentation and fluorescence traces in the NWB file. If False,
            neuropil traces are included in the main plane segmentation rather than the background plane segmentation.
        include_roi_centroids : bool, default: True
            Whether to include the ROI centroids on the PlaneSegmentation table.
            If there are a very large number of ROIs (such as in whole-brain recordings),
            you may wish to disable this for faster write speeds.
        include_roi_acceptance : bool, default: True
            Whether to include if the detected ROI was 'accepted' or 'rejected'.
            If there are a very large number of ROIs (such as in whole-brain recordings), you may wish to ddisable this for
            faster write speeds.
        mask_type : str, default: 'image'
            There are three types of ROI masks in NWB, 'image', 'pixel', and 'voxel'.

            * 'image' masks have the same shape as the reference images the segmentation was applied to, and weight each pixel
              by its contribution to the ROI (typically boolean, with 0 meaning 'not in the ROI').
            * 'pixel' masks are instead indexed by ROI, with the data at each index being the shape of the image by the number
              of pixels in each ROI.
            * 'voxel' masks are instead indexed by ROI, with the data at each index being the shape of the volume by the number
              of voxels in each ROI.

            Specify your choice between these two as mask_type='image', 'pixel', 'voxel', or None.
            If None, the mask information is not written to the NWB file.
        plane_segmentation_name : str, optional
            The name of the plane segmentation to be added.
        iterator_options : dict, optional
            The options to use when iterating over the image masks of the segmentation extractor.

        Returns
        -------

        """
        from ...tools.roiextractors import add_segmentation_to_nwbfile

        if stub_test:
            stub_frames = min([stub_frames, self.segmentation_extractor.get_num_frames()])
            segmentation_extractor = self.segmentation_extractor.frame_slice(start_frame=0, end_frame=stub_frames)
        else:
            segmentation_extractor = self.segmentation_extractor

        add_segmentation_to_nwbfile(
            segmentation_extractor=segmentation_extractor,
            nwbfile=nwbfile,
            metadata=metadata,
            include_background_segmentation=include_background_segmentation,
            include_roi_centroids=include_roi_centroids,
            include_roi_acceptance=include_roi_acceptance,
            mask_type=mask_type,
            plane_segmentation_name=plane_segmentation_name,
            iterator_options=iterator_options,
            compression_options=compression_options,  # TODO: remove completely after 10/1/2024; still passing for deprecation warning
        )



from pydantic import DirectoryPath

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface


class AlphaOmegaRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface class for converting AlphaOmega recording data.

    Uses the :py:class:`~spikeinterface.extractors.AlphaOmegaRecordingExtractor`.
    """

    display_name = "AlphaOmega Recording"
    associated_suffixes = (".mpx",)
    info = "Interface class for converting AlphaOmega recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"]["description"] = "Path to the folder of .mpx files."
        return source_schema

    def __init__(self, folder_path: DirectoryPath, verbose: bool = True, es_key: str = "ElectricalSeries"):
        """
        Load and prepare data for AlphaOmega.

        Parameters
        ----------
        folder_path: string or Path
            Path to the folder of .mpx files.
        verbose: boolean
            Allows verbose.
            Default is True.
        es_key: str, default: "ElectricalSeries"
        """
        super().__init__(folder_path=folder_path, stream_id="RAW", verbose=verbose, es_key=es_key)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        annotation = self.recording_extractor.neo_reader.raw_annotations
        metadata["NWBFile"].update(session_start_time=annotation["blocks"][0]["rec_datetime"])
        return metadata



from ._mock._mock_dataset_models import (
    mock_HDF5BackendConfiguration,
    mock_HDF5DatasetIOConfiguration,
    mock_ZarrBackendConfiguration,
    mock_ZarrDatasetIOConfiguration,
)
from .mock_files import generate_path_expander_demo_ibl
from .mock_interfaces import (
    MockBehaviorEventInterface,
    MockSpikeGLXNIDQInterface,
    MockRecordingInterface,
    MockImagingInterface,
    MockSortingInterface,
)
from .mock_ttl_signals import generate_mock_ttl_signal, regenerate_test_cases



import distutils.version
import uuid
import warnings
from copy import deepcopy
from pathlib import Path
from typing import Optional

import neo.io.baseio
import numpy as np
import pynwb
from pydantic import FilePath

from ..nwb_helpers import add_device_from_metadata

response_classes = dict(
    voltage_clamp=pynwb.icephys.VoltageClampSeries,
    current_clamp=pynwb.icephys.CurrentClampSeries,
    izero=pynwb.icephys.IZeroClampSeries,
)

stim_classes = dict(
    voltage_clamp=pynwb.icephys.VoltageClampStimulusSeries,
    current_clamp=pynwb.icephys.CurrentClampStimulusSeries,
)


# TODO - get electrodes metadata
def get_electrodes_metadata(neo_reader, electrodes_ids: list, block: int = 0) -> list:
    """
    Get electrodes metadata from Neo reader.

    The typical information we look for is the information accepted by pynwb.icephys.IntracellularElectrode:
      - name – the name of this electrode
      - device – the device that was used to record from this electrode
      - description – Recording description, description of electrode (e.g., whole-cell, sharp, etc.)
      - comment: Free-form text (can be from Methods)
      - slice – Information about slice used for recording.
      - seal – Information about seal used for recording.
      - location – Area, layer, comments on estimation, stereotaxis coordinates (if in vivo, etc.).
      - resistance – Electrode resistance COMMENT: unit: Ohm.
      - filtering – Electrode specific filtering.
      - initial_access_resistance – Initial access resistance.

    Parameters
    ----------
    neo_reader ([type]): Neo reader
    electrodes_ids (list): List of electrodes ids.
    block (int, optional): Block id. Defaults to 0.

    Returns
    -------
    list: List of dictionaries containing electrodes metadata.
    """
    return []


def get_number_of_electrodes(neo_reader) -> int:
    """Get number of electrodes from Neo reader."""
    # TODO - take in account the case with multiple streams.
    return len(neo_reader.header["signal_channels"])


def get_number_of_segments(neo_reader, block: int = 0) -> int:
    """Get number of segments from Neo reader."""
    return neo_reader.header["nb_segment"][block]


def get_command_traces(neo_reader, segment: int = 0, cmd_channel: int = 0) -> tuple[list, str, str]:
    """
    Get command traces (e.g. voltage clamp command traces).

    Parameters
    ----------
    neo_reader : neo.io.baseio
    segment : int, optional
        Defaults to 0.
    cmd_channel : int, optional
        ABF command channel (0 to 7). Defaults to 0.
    """
    try:
        traces, titles, units = neo_reader.read_raw_protocol()
        return traces[segment][cmd_channel], titles[segment][cmd_channel], units[segment][cmd_channel]
    except Exception as e:
        msg = ".\n\n WARNING - get_command_traces() only works for AxonIO interface."
        e.args = (str(e) + msg,)
        return e


def get_conversion_from_unit(unit: str) -> float:
    """
    Get conversion (to Volt or Ampere) from unit in string format.

    Parameters
    ----------
    unit (str): Unit as string. E.g. pA, mV, uV, etc...

    Returns
    -------
    float: conversion to Ampere or Volt
    """
    if unit in ["pA", "pV"]:
        conversion = 1e-12
    elif unit in ["nA", "nV"]:
        conversion = 1e-9
    elif unit in ["uA", "uV"]:
        conversion = 1e-6
    elif unit in ["mA", "mV"]:
        conversion = 1e-3
    elif unit in ["A", "V"]:
        conversion = 1.0
    else:
        conversion = 1.0
        warnings.warn("No valid units found for traces in the current file. Gain is set to 1, but this might be wrong.")
    return float(conversion)


def get_nwb_metadata(neo_reader, metadata: dict = None) -> dict:
    """
    Return default metadata for all recording fields.

    Parameters
    ----------
    neo_reader: Neo reader object
    metadata: dict, optional
        Metadata info for constructing the nwb file.
    """
    metadata = dict(
        NWBFile=dict(
            session_description="Auto-generated by NwbRecordingExtractor without description.",
            identifier=str(uuid.uuid4()),
        ),
        Icephys=dict(Device=[dict(name="Device", description="no description")]),
    )
    return metadata


def add_icephys_electrode(neo_reader, nwbfile, metadata: dict = None):
    """
    Add icephys electrodes to nwbfile object.

    Will always ensure nwbfile has at least one icephys electrode.
    Will auto-generate a linked device if the specified name does not exist in the nwbfile.

    Parameters
    ----------
    neo_reader : neo.io.baseio
    nwbfile : NWBFile
        NWBFile object to add the icephys electrode to.
    metadata : dict, optional
        Metadata info for constructing the nwb file.
        Should be of the format::

            metadata['Icephys']['Electrodes'] = [
                {
                    'name': my_name,
                    'description': my_description,
                    'device_name': my_device_name
                },
                ...
            ]
    """
    metadata_copy = deepcopy(metadata) if metadata is not None else dict()

    assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    if len(nwbfile.devices) == 0:
        warnings.warn("When adding Icephys Electrode, no Devices were found on nwbfile. Creating a Device now...")
        add_device_from_metadata(nwbfile=nwbfile, modality="Icephys", metadata=metadata_copy)

    if "Icephys" not in metadata_copy:
        metadata_copy["Icephys"] = dict()

    defaults = [
        dict(
            name=f"icephys_electrode_{elec_id}",
            description="no description",
            device_name=[i.name for i in nwbfile.devices.values()][0],
        )
        for elec_id in range(get_number_of_electrodes(neo_reader))
    ]

    if "Electrodes" not in metadata_copy["Icephys"] or len(metadata_copy["Icephys"]["Electrodes"]) == 0:
        metadata_copy["Icephys"]["Electrodes"] = defaults

    assert all(
        [isinstance(x, dict) for x in metadata_copy["Icephys"]["Electrodes"]]
    ), "Expected metadata['Icephys']['Electrodes'] to be a list of dictionaries!"

    # Create Icephys electrode from metadata
    for elec in metadata_copy["Icephys"]["Electrodes"]:
        if elec.get("name", defaults[0]["name"]) not in nwbfile.icephys_electrodes:
            device_name = elec.pop("device_name", None) or elec.pop("device", defaults[0]["device_name"])
            # elec.pop("device_name", 0)
            if device_name not in nwbfile.devices:
                new_device_metadata = dict(Ecephys=dict(Device=[dict(name=device_name)]))
                add_device_from_metadata(nwbfile, modality="Icephys", metadata=new_device_metadata)
                warnings.warn(
                    f"Device '{device_name}' not detected in "
                    "attempted link to icephys electrode! Automatically generating."
                )
            electrode_kwargs = elec
            electrode_kwargs.update(
                name=elec.get("name", defaults[0]["name"]),
                description=elec.get("description", defaults[0]["description"]),
                device=nwbfile.devices[device_name],
            )
            nwbfile.create_icephys_electrode(**electrode_kwargs)


def add_icephys_recordings(
    neo_reader,
    nwbfile: pynwb.NWBFile,
    metadata: dict = None,
    icephys_experiment_type: str = "voltage_clamp",
    stimulus_type: str = "not described",
    skip_electrodes: tuple[int] = (),
    compression: Optional[str] = None,  # TODO: remove completely after 10/1/2024
):
    """
    Add icephys recordings (stimulus/response pairs) to nwbfile object.

    Parameters
    ----------
    neo_reader : neo.io.baseio
    nwbfile : NWBFile
    metadata : dict, optional
    icephys_experiment_type : {'voltage_clamp', 'current_clamp', 'izero'}
        Type of icephys recording.
    stimulus_type : str, default: 'not described'
    skip_electrodes : tuple, default: ()
        Electrode IDs to skip.
    """
    # TODO: remove completely after 10/1/2024
    if compression is not None:
        warn(
            message=(
                "Specifying compression methods and their options at the level of tool functions has been deprecated. "
                "Please use the `configure_backend` tool function for this purpose."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

    n_segments = get_number_of_segments(neo_reader, block=0)

    # Check for protocol data (only ABF2), necessary for stimuli data
    if neo_reader._axon_info["fFileVersionNumber"] < 2:
        n_commands = 0
        warnings.warn(
            f"Protocol section is only present in ABF2 files. {neo_reader.filename} has version "
            f"{neo_reader._axon_info['fFileVersionNumber']}. Saving experiment as 'i_zero'..."
        )
    else:
        protocol = neo_reader.read_raw_protocol()
        n_commands = len(protocol[0])

    if n_commands == 0:
        icephys_experiment_type = "izero"
        warnings.warn(
            f"No command data found by neo reader in file {neo_reader.filename}. Saving experiment as 'i_zero'..."
        )
    else:
        assert (
            n_commands == n_segments
        ), f"File contains inconsistent number of segments ({n_segments}) and commands ({n_commands})"

    assert icephys_experiment_type in ["voltage_clamp", "current_clamp", "izero"], (
        f"'icephys_experiment_type' should be 'voltage_clamp', 'current_clamp' or 'izero', but received value "
        f"{icephys_experiment_type}"
    )

    # Check and auto-create electrodes, in case they don't exist yet in nwbfile
    if len(nwbfile.icephys_electrodes) == 0:
        warnings.warn(
            "When adding Icephys Recording, no Icephys Electrodes were found on nwbfile. Creating Electrodes now..."
        )
        add_icephys_electrode(
            neo_reader=neo_reader,
            nwbfile=nwbfile,
            metadata=metadata,
        )

    if getattr(nwbfile, "intracellular_recordings", None):
        ri = max(nwbfile.intracellular_recordings["responses"].index)
    else:
        ri = -1

    if getattr(nwbfile, "icephys_simultaneous_recordings", None):
        simultaneous_recordings_offset = len(nwbfile.icephys_simultaneous_recordings)
    else:
        simultaneous_recordings_offset = 0

    if getattr(nwbfile, "icephys_sequential_recordings", None):
        sessions_offset = len(nwbfile.icephys_sequential_recordings)
    else:
        sessions_offset = 0

    relative_session_start_time = deepcopy(
        metadata["Icephys"]["Sessions"][sessions_offset]["relative_session_start_time"]
    )
    session_stimulus_type = deepcopy(metadata["Icephys"]["Sessions"][sessions_offset]["stimulus_type"])

    # Sequential icephys recordings
    simultaneous_recordings = list()
    for si in range(n_segments):
        # Parallel icephys recordings
        recordings = list()
        for ei, electrode in enumerate(
            list(nwbfile.icephys_electrodes.values())[: len(neo_reader.header["signal_channels"]["units"])]
        ):
            if ei in skip_electrodes:
                continue
            # Starting time is the signal starting time within .abf file + time
            # relative to first session (first .abf file)
            ri += 1
            starting_time = neo_reader.get_signal_t_start(block_index=0, seg_index=si)
            starting_time = starting_time + relative_session_start_time

            sampling_rate = neo_reader.get_signal_sampling_rate()
            response_unit = neo_reader.header["signal_channels"]["units"][ei]
            response_conversion = get_conversion_from_unit(unit=response_unit)
            response_gain = neo_reader.header["signal_channels"]["gain"][ei]
            response_name = f"{icephys_experiment_type}-response-{si + 1 + simultaneous_recordings_offset:02}-ch-{ei}"

            response = response_classes[icephys_experiment_type](
                name=response_name,
                description=f"Response to: {session_stimulus_type}",
                electrode=electrode,
                data=neo_reader.get_analogsignal_chunk(block_index=0, seg_index=si, channel_indexes=ei),
                starting_time=starting_time,
                rate=sampling_rate,
                conversion=response_conversion * response_gain,
                gain=np.nan,
            )
            if icephys_experiment_type != "izero":
                stim_unit = protocol[2][ei]
                stim_conversion = get_conversion_from_unit(unit=stim_unit)
                stimulus = stim_classes[icephys_experiment_type](
                    name=f"stimulus-{si + 1 + simultaneous_recordings_offset:02}-ch-{ei}",
                    description=f"Stim type: {session_stimulus_type}",
                    electrode=electrode,
                    data=protocol[0][si][ei],
                    rate=sampling_rate,
                    starting_time=starting_time,
                    conversion=stim_conversion,
                    gain=np.nan,
                )
                icephys_recording = nwbfile.add_intracellular_recording(
                    electrode=electrode, response=response, stimulus=stimulus
                )
            else:
                icephys_recording = nwbfile.add_intracellular_recording(electrode=electrode, response=response)

            recordings.append(icephys_recording)

        sim_rec = nwbfile.add_icephys_simultaneous_recording(recordings=recordings)
        simultaneous_recordings.append(sim_rec)

    nwbfile.add_icephys_sequential_recording(
        simultaneous_recordings=simultaneous_recordings, stimulus_type=stimulus_type
    )

    # TODO
    # # Add a list of sequential recordings table indices as a repetition
    # run_index = nwbfile.add_icephys_repetition(
    #     sequential_recordings=[
    #         seq_rec,
    #     ]
    # )

    # # Add a list of repetition table indices as a experimental condition
    # nwbfile.add_icephys_experimental_condition(
    #     repetitions=[
    #         run_index,
    #     ]
    # )


def add_neo_to_nwb(
    neo_reader,
    nwbfile: pynwb.NWBFile,
    metadata: dict = None,
    compression: Optional[str] = None,  # TODO: remove completely after 10/1/2024
    icephys_experiment_type: str = "voltage_clamp",
    stimulus_type: Optional[str] = None,
    skip_electrodes: tuple[int] = (),
):
    """
    Auxiliary static method for nwbextractor.

    Adds all recording related information from recording object and metadata to the nwbfile object.

    Parameters
    ----------
    neo_reader: Neo reader object
    nwbfile: NWBFile
        nwb file to which the recording information is to be added
    metadata: dict
        metadata info for constructing the nwb file (optional).
        Check the auxiliary function docstrings for more information
        about metadata format.
    icephys_experiment_type: str (optional)
        Type of Icephys experiment. Allowed types are: 'voltage_clamp', 'current_clamp' and 'izero'.
        If no value is passed, 'voltage_clamp' is used as default.
    stimulus_type: str, optional
    skip_electrodes: tuple, optional
        Electrode IDs to skip. Defaults to ().
    """
    assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    # TODO: remove completely after 10/1/2024
    if compression is not None:
        warn(
            message=(
                "Specifying compression methods and their options at the level of tool functions has been deprecated. "
                "Please use the `configure_backend` tool function for this purpose."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

    add_device_from_metadata(nwbfile=nwbfile, modality="Icephys", metadata=metadata)

    add_icephys_electrode(
        neo_reader=neo_reader,
        nwbfile=nwbfile,
        metadata=metadata,
    )

    add_icephys_recordings(
        neo_reader=neo_reader,
        nwbfile=nwbfile,
        metadata=metadata,
        icephys_experiment_type=icephys_experiment_type,
        stimulus_type=stimulus_type,
        skip_electrodes=skip_electrodes,
    )


def write_neo_to_nwb(
    neo_reader: neo.io.baseio.BaseIO,
    save_path: Optional[FilePath] = None,  # pragma: no cover
    overwrite: bool = False,
    nwbfile=None,
    metadata: dict = None,
    compression: Optional[str] = None,  # TODO: remove completely after 10/1/2024
    icephys_experiment_type: Optional[str] = None,
    stimulus_type: Optional[str] = None,
    skip_electrodes: Optional[tuple] = (),
):
    """
    Primary method for writing a Neo reader object to an NWBFile.

    Parameters
    ----------
    neo_reader: Neo reader
    save_path: PathType
        Required if an nwbfile is not passed. Must be the path to the nwbfile
        being appended, otherwise one is created and written.
    overwrite: bool
        If using save_path, whether to overwrite the NWBFile if it already exists.
    nwbfile: NWBFile
        Required if a save_path is not specified. If passed, this function
        will fill the relevant fields within the nwbfile.
    metadata: dict
        metadata info for constructing the nwb file (optional). Should be of the format::

            metadata['Ecephys'] = {}

        with keys of the forms::

            metadata['Ecephys']['Device'] = [
                {
                    'name': my_name,
                    'description': my_description
                },
                ...
            ]
            metadata['Ecephys']['ElectrodeGroup'] = [
                {
                    'name': my_name,
                    'description': my_description,
                    'location': electrode_location,
                    'device': my_device_name
                },
                ...
            ]
            metadata['Ecephys']['Electrodes'] = [
                {
                    'name': my_name,
                    'description': my_description
                },
                ...
            ]
            metadata['Ecephys']['ElectricalSeries'] = {
                'name': my_name,
                'description': my_description
            }

        Note that data intended to be added to the electrodes table of the NWBFile should be set as channel
        properties in the RecordingExtractor object.
    compression: str (optional, defaults to "gzip")
        Type of compression to use. Valid types are "gzip" and "lzf".
        Set to None to disable all compression.
    icephys_experiment_type: str (optional)
        Type of Icephys experiment. Allowed types are: 'voltage_clamp', 'current_clamp' and 'izero'.
        If no value is passed, 'voltage_clamp' is used as default.
    stimulus_type: str, optional
    skip_electrodes: tuple, optional
        Electrode IDs to skip. Defaults to ().
    """
    if nwbfile is not None:
        assert isinstance(nwbfile, pynwb.NWBFile), "'nwbfile' should be of type pynwb.NWBFile"

    assert (
        distutils.version.LooseVersion(pynwb.__version__) >= "1.3.3"
    ), "'write_neo_to_nwb' not supported for version < 1.3.3. Run pip install --upgrade pynwb"

    assert save_path is None or nwbfile is None, "Either pass a save_path location, or nwbfile object, but not both!"

    # TODO: remove completely after 10/1/2024
    if compression is not None:
        warn(
            message=(
                "Specifying compression methods and their options at the level of tool functions has been deprecated. "
                "Please use the `configure_backend` tool function for this purpose."
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

    if metadata is None:
        metadata = get_nwb_metadata(neo_reader=neo_reader)

    kwargs = dict(
        neo_reader=neo_reader,
        metadata=metadata,
        icephys_experiment_type=icephys_experiment_type,
        stimulus_type=stimulus_type,
        skip_electrodes=skip_electrodes,
    )
    if nwbfile is None:
        if Path(save_path).is_file() and not overwrite:
            read_mode = "r+"
        else:
            read_mode = "w"

        with pynwb.NWBHDF5IO(str(save_path), mode=read_mode) as io:
            if read_mode == "r+":
                nwbfile = io.read()
            else:
                nwbfile_kwargs = dict(
                    session_description="Auto-generated by NwbRecordingExtractor without description.",
                    identifier=str(uuid.uuid4()),
                )
                if metadata is not None and "NWBFile" in metadata:
                    nwbfile_kwargs.update(metadata["NWBFile"])
                nwbfile = pynwb.NWBFile(**nwbfile_kwargs)

            add_neo_to_nwb(nwbfile=nwbfile, **kwargs)
            io.write(nwbfile)
    else:
        add_neo_to_nwb(nwbfile=nwbfile, **kwargs)



"""Collection of helper functions for assessing and performing automated data transfers."""

import json
import re
from pathlib import Path
from time import sleep, time
from typing import Union

from pydantic import DirectoryPath
from tqdm import tqdm

from ..importing import is_package_installed
from ..processes import deploy_process


def get_globus_dataset_content_sizes(
    globus_endpoint_id: str, path: str, recursive: bool = True, timeout: float = 120.0
) -> dict[str, int]:  # pragma: no cover
    """
    May require external login via 'globus login' from CLI.

    Returns dictionary whose keys are file names and values are sizes in bytes.
    """
    assert is_package_installed(package_name="globus_cli"), "You must install the globus CLI (pip install globus-cli)!"

    recursive_flag = " --recursive" if recursive else ""
    contents = json.loads(
        deploy_process(
            command=f"globus ls -Fjson {globus_endpoint_id}:{path}{recursive_flag}", catch_output=True, timeout=timeout
        )
    )
    files_and_sizes = {item["name"]: item["size"] for item in contents["DATA"] if item["type"] == "file"}
    return files_and_sizes


def transfer_globus_content(
    source_endpoint_id: str,
    source_files: Union[str, list[list[str]]],
    destination_endpoint_id: str,
    destination_folder: DirectoryPath,
    display_progress: bool = True,
    progress_update_rate: float = 60.0,
    progress_update_timeout: float = 600.0,
) -> tuple[bool, list[str]]:  # pragma: no cover
    """
    Track progress for transferring content from source_endpoint_id to destination_endpoint_id:destination_folder.

    Parameters
    ----------
    source_endpoint_id : str
        Source Globus ID.
    source_files : string, or list of strings, or list of lists of strings
        A string path or list-of-lists of string paths of files to transfer from the source_endpoint_id.
        If using a nested list, the outer level indicates which requests will be batched together.
        If using a nested list, all items in a single batch level must be from the same common directory.

        It is recommended to transfer the largest file(s) with minimal batching,
        and to batch a large number of very small files together.

        It is also generally recommended to submit up to 3 simultaneous transfer,
        *i.e.*, `source_files` is recommended to have at most 3 items all of similar total byte size.
    destination_endpoint_id : str
        Destination Globus ID.
    destination_folder : FolderPathType
        Absolute path to a local folder where all content will be transferred to.
    display_progress : bool, default: True
        Whether to display the transfer as progress bars using `tqdm`.
    progress_update_rate : float, default: 60.0
        How frequently (in seconds) to update the progress bar display tracking the data transfer.
    progress_update_timeout : float, default: 600.0
        Maximum amount of time to monitor the transfer progress.
        You may wish to set this to be longer when transferring very large files.

    Returns
    -------
    success : bool
        Returns the total status of all transfers when they either finish or the progress tracking times out.
    task_ids : list of strings
        List of the task IDs submitted to globus, if further information is needed to reestablish tracking or terminate.
    """

    def _submit_transfer_request(
        source_endpoint_id: str,
        source_files: Union[str, list[list[str]]],
        destination_endpoint_id: str,
        destination_folder_path: Path,
    ) -> dict[str, int]:
        """Send transfer request to Globus."""
        folder_content_sizes = dict()
        task_total_sizes = dict()
        for j, batched_source_files in enumerate(source_files):
            paths_file = destination_folder_path / f"paths_{j}.txt"
            # .as_posix() to ensure correct string form Globus expects
            # ':' replacement for Windows drives
            source_folder = Path(batched_source_files[0]).parent.as_posix().replace(":", "")
            destination_folder_name = destination_folder_path.as_posix().replace(":", "")
            with open(file=paths_file, mode="w") as f:
                for source_file in batched_source_files:
                    file_name = Path(source_file).name
                    f.write(f"{file_name} {file_name}\n")

            transfer_command = (
                "globus transfer "
                f"{source_endpoint_id}:{source_folder} {destination_endpoint_id}:/{destination_folder_name} "
                f"--batch {paths_file}"
            )
            transfer_message = deploy_process(
                command=transfer_command,
                catch_output=True,
            )
            task_id = re.findall(
                pattern=(
                    "^Message: The transfer has been accepted and a task has been created and queued for "
                    "execution\nTask ID: (.+)$"
                ),
                string=transfer_message,
            )
            paths_file.unlink()
            assert task_id, f"Transfer submission failed! Globus output:\n{transfer_message}."

            if source_folder not in folder_content_sizes:
                contents = get_globus_dataset_content_sizes(globus_endpoint_id=source_endpoint_id, path=source_folder)
                folder_content_sizes.update({source_folder: contents})
            task_total_sizes.update(
                {
                    task_id[0]: sum(
                        [
                            folder_content_sizes[source_folder][Path(source_file).name]
                            for source_file in batched_source_files
                        ]
                    )
                }
            )
        return task_total_sizes

    def _track_transfer(
        task_total_sizes: dict[str, int],
        display_progress: bool = True,
        progress_update_rate: float = 60.0,
        progress_update_timeout: float = 600.0,
    ) -> bool:
        """Track the progress of transfers."""
        if display_progress:
            all_pbars = [
                tqdm(desc=f"Transferring batch #{j}...", total=total_size, position=j, leave=True)
                for j, total_size in enumerate(task_total_sizes.values(), start=1)
            ]
        all_status = [False for _ in task_total_sizes]
        success = all(all_status)
        time_so_far = 0.0
        start_time = time()
        while not success and time_so_far <= progress_update_timeout:
            time_so_far = time() - start_time
            for j, (task_id, task_total_size) in enumerate(task_total_sizes.items()):
                task_update = deploy_process(f"globus task show {task_id} -Fjson", catch_output=True)
                task_message = json.loads(task_update)
                all_status[j] = task_message["status"] == "SUCCEEDED"
                assert (
                    all_status[j] != "OK" or all_status[j] != "SUCCEEDED"
                ), f"Something went wrong with the transfer! Please manually inspect the task with ID '{task_id}'."
                if display_progress:
                    all_pbars[j].update(n=task_message["bytes_transferred"] - all_pbars[j].n)
            success = all(all_status)
            if not success:
                sleep(progress_update_rate)
        return success

    source_files = [[source_files]] if isinstance(source_files, str) else source_files
    destination_folder_path = Path(destination_folder)
    destination_folder_path.mkdir(exist_ok=True)
    # assertion check is ensure the logical iteration does not occur over the individual string values
    assert (
        isinstance(source_files, list) and source_files and isinstance(source_files[0], list) and source_files[0]
    ), "'source_files' must be a non-empty nested list-of-lists to indicate batched transfers!"
    assert destination_folder_path.is_absolute(), (
        "The 'destination_folder' must be an absolute path to resolve ambiguity with relative Globus head "
        "as well as avoiding Globus access permissions to a personal relative path!"
    )

    task_total_sizes = _submit_transfer_request(
        source_endpoint_id=source_endpoint_id,
        source_files=source_files,
        destination_endpoint_id=destination_endpoint_id,
        destination_folder_path=destination_folder_path,
    )
    success = _track_transfer(
        task_total_sizes=task_total_sizes,
        display_progress=display_progress,
        progress_update_rate=progress_update_rate,
        progress_update_timeout=progress_update_timeout,
    )
    return success, list(task_total_sizes)



from typing import Optional

from pydantic import DirectoryPath, validate_call
from pynwb import NWBFile

from ... import MiniscopeBehaviorInterface, MiniscopeImagingInterface
from ....nwbconverter import NWBConverter
from ....tools.nwb_helpers import make_or_load_nwbfile
from ....utils import get_schema_from_method_signature


class MiniscopeConverter(NWBConverter):
    """Primary conversion class for handling Miniscope data streams."""

    display_name = "Miniscope Imaging and Video"
    keywords = MiniscopeImagingInterface.keywords + MiniscopeBehaviorInterface.keywords
    associated_suffixes = MiniscopeImagingInterface.associated_suffixes + MiniscopeBehaviorInterface.associated_suffixes
    info = "Converter for handling both imaging and video recordings from Miniscope."

    @classmethod
    def get_source_schema(cls):
        source_schema = get_schema_from_method_signature(cls)
        source_schema["properties"]["folder_path"]["description"] = "The path to the main Miniscope folder."
        return source_schema

    @validate_call
    def __init__(self, folder_path: DirectoryPath, verbose: bool = True):
        """
        Initializes the data interfaces for the Miniscope recording and behavioral data stream.

        The main Miniscope folder is expected to contain both data streams organized as follows::

            C6-J588_Disc5/ (main folder)
            ├── 15_03_28/ (subfolder corresponding to the recording time)
            │   ├── Miniscope/ (subfolder containing the microscope video stream)
            │   │   ├── 0.avi (microscope video)
            │   │   ├── metaData.json (metadata for the microscope device)
            │   │   └── timeStamps.csv (timing of this video stream)
            │   ├── BehavCam_2/ (subfolder containing the behavioral video stream)
            │   │   ├── 0.avi (bevavioral video)
            │   │   ├── metaData.json (metadata for the behavioral camera)
            │   │   └── timeStamps.csv (timing of this video stream)
            │   └── metaData.json (metadata for the recording, such as the start time)
            ├── 15_06_28/
            │   ├── Miniscope/
            │   ├── BehavCam_2/
            │   └── metaData.json
            └── 15_12_28/

        Parameters
        ----------
        folder_path : FolderPathType
            The path to the main Miniscope folder.
        verbose : bool, default: True
            Controls verbosity.
        """
        self.verbose = verbose
        self.data_interface_objects = dict(
            MiniscopeImaging=MiniscopeImagingInterface(folder_path=folder_path),
            MiniscopeBehavCam=MiniscopeBehaviorInterface(folder_path=folder_path),
        )

    def get_conversion_options_schema(self) -> dict:
        return self.data_interface_objects["MiniscopeImaging"].get_conversion_options_schema()

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata,
        stub_test: bool = False,
        stub_frames: int = 100,
    ):
        self.data_interface_objects["MiniscopeImaging"].add_to_nwbfile(
            nwbfile=nwbfile,
            metadata=metadata,
            stub_test=stub_test,
            stub_frames=stub_frames,
        )
        self.data_interface_objects["MiniscopeBehavCam"].add_to_nwbfile(
            nwbfile=nwbfile,
            metadata=metadata,
        )

    def run_conversion(
        self,
        nwbfile_path: Optional[str] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        stub_test: bool = False,
        stub_frames: int = 100,
    ) -> None:
        if metadata is None:
            metadata = self.get_metadata()

        self.validate_metadata(metadata=metadata)

        self.temporally_align_data_interfaces()

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
        ) as nwbfile_out:
            self.add_to_nwbfile(nwbfile=nwbfile_out, metadata=metadata, stub_test=stub_test, stub_frames=stub_frames)



import re
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Optional

import numpy as np
from pydantic import FilePath, validate_call
from pynwb import NWBFile

from ....basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ....tools import get_module
from ....utils import (
    DeepDict,
    calculate_regular_series_rate,
    get_base_schema,
)


class LightningPoseDataInterface(BaseTemporalAlignmentInterface):
    """Data interface for Lightning Pose datasets."""

    display_name = "Lightning Pose"
    keywords = ("pose estimation", "video")
    associated_suffixes = (".csv", ".mp4")
    info = "Interface for handling a single stream of lightning pose data."

    def get_metadata_schema(self) -> dict:
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Behavior"] = get_base_schema(tag="Behavior")

        metadata_schema["properties"]["Behavior"].update(
            required=["PoseEstimation"],
            properties=dict(
                PoseEstimation=dict(
                    type="object",
                    required=["name"],
                    properties=dict(
                        name=dict(type="string", default="PoseEstimation"),
                        description=dict(type="string"),
                        scorer=dict(type="string"),
                        source_software=dict(type="string", default="LightningPose"),
                    ),
                    patternProperties={
                        "^(?!(name|description|scorer|source_software)$)[a-zA-Z0-9_]+$": dict(
                            title="PoseEstimationSeries",
                            type="object",
                            properties=dict(name=dict(type="string"), description=dict(type="string")),
                            minProperties=1,
                            additionalProperties=False,
                        )
                    },
                    minProperties=2,
                    additionalProperties=False,
                )
            ),
        )

        return metadata_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        original_video_file_path: FilePath,
        labeled_video_file_path: Optional[FilePath] = None,
        verbose: bool = True,
    ):
        """
        Interface for writing pose estimation data from the Lightning Pose algorithm.

        Parameters
        ----------
        file_path : FilePath
            Path to the .csv file that contains the predictions from Lightning Pose.
        original_video_file_path : FilePath
            Path to the original video file (.mp4).
        labeled_video_file_path : a string or a path, optional
            Path to the labeled video file (.mp4).
        verbose : bool, default: True
            controls verbosity. ``True`` by default.
        """
        from neuroconv.datainterfaces.behavior.video.video_utils import (
            VideoCaptureContext,
        )

        self._vc = VideoCaptureContext

        self.file_path = Path(file_path)
        assert self.file_path.exists(), f"The file '{self.file_path}' does not exist."
        self.original_video_file_path = Path(original_video_file_path)
        assert (
            self.original_video_file_path.exists()
        ), f"The original video file '{self.original_video_file_path}' does not exist."

        super().__init__(
            verbose,
            file_path=file_path,
            original_video_file_path=original_video_file_path,
            labeled_video_file_path=labeled_video_file_path,
        )

        # dimension is width by height
        self.dimension = self._get_original_video_shape()

        pose_estimation_data = self._load_source_data()
        _, self.scorer_name = pose_estimation_data.columns.get_level_values(0).drop_duplicates()
        self.pose_estimation_data = pose_estimation_data[self.scorer_name]
        self.keypoint_names = self.pose_estimation_data.columns.get_level_values(0).drop_duplicates().tolist()

        self._times = None

    def _load_source_data(self):
        import pandas as pd

        # The order of the header is "scorer", "bodyparts", "coords"
        pose_estimation_data = pd.read_csv(self.file_path, header=[0, 1, 2])
        return pose_estimation_data

    def _get_original_video_shape(self) -> tuple[int, int]:
        with self._vc(file_path=str(self.original_video_file_path)) as video:
            video_shape = video.get_frame_shape()
        # image size of the original video is in height x width
        return video_shape[0], video_shape[1]

    def get_original_timestamps(self, stub_test: bool = False) -> np.ndarray:
        max_frames = 10 if stub_test else None
        with self._vc(file_path=str(self.original_video_file_path)) as video:
            timestamps = video.get_video_timestamps(max_frames=max_frames)
        return timestamps

    def get_timestamps(self, stub_test: bool = False) -> np.ndarray:
        max_frames = 10 if stub_test else None
        if self._times is None:
            return self.get_original_timestamps(stub_test=stub_test)

        timestamps = self._times if not stub_test else self._times[:max_frames]
        return timestamps

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        self._times = aligned_timestamps

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        # Update the session start time if folder structure is saved in the format: YYYY-MM-DD/HH-MM-SS
        pattern = r"(?P<date_time>\d{4}-\d{2}-\d{2}/\d{2}-\d{2}-\d{2})"
        # Convert the file path parts to a string with forward slashes
        file_path = "/".join(self.file_path.parts)
        match = re.search(pattern, file_path)
        if match and "session_start_time" not in metadata["NWBFile"]:
            datetime_str = match.group("date_time")
            session_start_time = datetime.strptime(datetime_str, "%Y-%m-%d/%H-%M-%S")
            metadata["NWBFile"].update(session_start_time=session_start_time)

        metadata["Behavior"]["PoseEstimation"].update(
            name="PoseEstimation",
            description="Contains the pose estimation series for each keypoint.",
            scorer=self.scorer_name,
            source_software="LightningPose",
        )
        for keypoint_name in self.keypoint_names:
            keypoint_name_without_spaces = keypoint_name.replace(" ", "")
            pose_estimation_series_metadata = {
                keypoint_name: dict(
                    name=f"PoseEstimationSeries{keypoint_name_without_spaces}",
                    description=f"The estimated position (x, y) of {keypoint_name} over time.",
                )
            }
            metadata["Behavior"]["PoseEstimation"].update(pose_estimation_series_metadata)

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        reference_frame: Optional[str] = None,
        confidence_definition: Optional[str] = None,
        stub_test: Optional[bool] = False,
    ) -> None:
        """
        Add the pose estimation data to the nwbfile.

        Parameters
        ----------
        nwbfile : NWBFile
            The nwbfile to which the pose estimation data is added.
        metadata : dict, optional
            The metadata for the pose estimation data.
        reference_frame : str, optional
            The description defining what the (0, 0) coordinate corresponds to.
        confidence_definition : str, optional
            The description of how the confidence was computed, e.g., 'Softmax output of the deep neural network'.
        stub_test : bool, default: False
        """
        from ndx_pose import PoseEstimation, PoseEstimationSeries

        metadata_copy = deepcopy(metadata)

        # The parameters for the pose estimation container
        pose_estimation_metadata = metadata_copy["Behavior"]["PoseEstimation"]

        behavior = get_module(nwbfile, "behavior")

        pose_estimation_name = pose_estimation_metadata["name"]
        if pose_estimation_name in behavior.data_interfaces:
            raise ValueError(f"The nwbfile already contains a data interface with the name '{pose_estimation_name}'.")

        if "Videos" not in metadata_copy["Behavior"]:
            original_video_name = str(self.original_video_file_path)
        else:
            original_video_name = metadata_copy["Behavior"]["Videos"][0]["name"]

        pose_estimation_kwargs = dict(
            name=pose_estimation_metadata["name"],
            description=pose_estimation_metadata["description"],
            source_software=pose_estimation_metadata["source_software"],
            scorer=pose_estimation_metadata["scorer"],
            original_videos=[original_video_name],
            dimensions=[self.dimension],
        )

        pose_estimation_data = self.pose_estimation_data if not stub_test else self.pose_estimation_data.head(n=10)
        timestamps = self.get_timestamps(stub_test=stub_test)
        rate = calculate_regular_series_rate(series=timestamps)
        if rate:
            pose_estimation_series_kwargs = dict(rate=rate, starting_time=timestamps[0])
        else:
            assert len(timestamps) == len(
                pose_estimation_data
            ), f"The length of timestamps ({len(timestamps)}) and pose estimation data ({len(pose_estimation_data)}) must be equal."
            pose_estimation_series_kwargs = dict(timestamps=timestamps)

        pose_estimation_series = []
        for keypoint_name in self.keypoint_names:
            pose_estimation_series_data = pose_estimation_data[keypoint_name]
            keypoint_name_without_spaces = keypoint_name.replace(" ", "")

            pose_estimation_series_kwargs.update(
                name=pose_estimation_metadata[keypoint_name_without_spaces]["name"],
                description=pose_estimation_metadata[keypoint_name_without_spaces]["description"],
                data=pose_estimation_series_data[["x", "y"]].values,
                confidence=pose_estimation_series_data["likelihood"].values,
                reference_frame=reference_frame or "(0,0) is unknown.",
                unit="px",
            )

            if confidence_definition:
                pose_estimation_series_kwargs.update(confidence_definition=confidence_definition)

            pose_estimation_series.append(PoseEstimationSeries(**pose_estimation_series_kwargs))

        pose_estimation_kwargs.update(
            pose_estimation_series=pose_estimation_series,
        )

        if self.source_data["labeled_video_file_path"]:
            if "Videos" not in metadata_copy["Behavior"]:
                labeled_video_name = str(self.source_data["labeled_video_file_path"])
            else:
                labeled_video_name = metadata_copy["Behavior"]["Videos"][1]["name"]

            pose_estimation_kwargs.update(labeled_videos=[labeled_video_name])

        # Create the container for pose estimation
        pose_estimation = PoseEstimation(**pose_estimation_kwargs)

        behavior.add(pose_estimation)



"""Tool functions related to imports."""

import sys
from importlib import import_module
from importlib.metadata import version as importlib_version
from importlib.util import find_spec
from platform import processor, python_version
from types import ModuleType
from typing import Optional, Union

from packaging import version


def get_package_version(name: str) -> version.Version:
    """
    Retrieve the version of a package.

    Parameters
    ----------
    name : str
        Name of package.

    Returns
    -------
    version : Version
        The package version as an object from packaging.version.Version.
    """
    package_version = importlib_version(name)
    return version.parse(package_version)


def is_package_installed(package_name: str) -> bool:
    """
    Check if the given package is installed on the system.

    Used for lazy imports.
    """
    return find_spec(name=package_name) is not None


def get_package(
    package_name: str,
    installation_instructions: Optional[str] = None,
    excluded_python_versions: Optional[list[str]] = None,
    excluded_platforms_and_python_versions: Optional[dict[str, Union[list[str], dict[str, list[str]]]]] = None,
) -> ModuleType:
    """
    Check if package is installed and return module if so.

    Otherwise, raise informative error describing how to perform the installation.
    Inspired by https://docs.python.org/3/library/importlib.html#checking-if-a-module-can-be-imported.

    Parameters
    ----------
    package_name : str
        Name of the package to be imported.
    installation_instructions : str, optional
        String describing the source, options, and alias of package name (if needed) for installation.
        For example,
            >>> installation_source = "conda install -c conda-forge my-package-name"
        Defaults to f"pip install {package_name}".
    excluded_python_versions : list of strs, optional
        If a given package has no distribution available for a certain Python version, it can be excluded by this
        import across all platforms. If you wish to be more specific about combinations of platforms and versions,
        use the 'excluded_platforms_and_python_versions' keyword argument instead.
        Allows all Python versions by default.
    excluded_platforms_and_python_versions : dict, optional
        Mapping of string platform names to a list of string versions.
        Valid platform strings are: ["linux", "win32", "darwin"] or any other variant used by sys.platform

        In case some combinations of platforms or Python versions are not allowed for the given package,
        specify this dictionary to raise a more specific error to that issue.

        For example, `excluded_platforms_and_python_versions = dict(darwin=["3.7"])` will raise an
        informative error when running on macOS with Python version 3.7.

        This also applies to specific architectures of platforms, such as
        `excluded_platforms_and_python_versions = dict(darwin=dict(arm=["3.7"]))` to exclude a specific Python
        version for M1 Macs.

        Allows all platforms and Python versions by default.

    Raises
    ------
    ModuleNotFoundError
    """
    installation_instructions = installation_instructions or f"pip install {package_name}"
    excluded_python_versions = excluded_python_versions or list()
    excluded_platforms_and_python_versions = excluded_platforms_and_python_versions or dict()

    python_minor_version = version.parse(python_version()).minor
    excluded_python_minor_versions = [
        version.parse(excluded_version).minor for excluded_version in excluded_python_versions
    ]
    if python_minor_version in excluded_python_minor_versions:
        raise ModuleNotFoundError(
            f"\nThe package '{package_name}' is not available for Python version 3.{python_minor_version}!"
        )

    # Specific architecture of specific platform is specified
    if isinstance(excluded_platforms_and_python_versions.get(sys.platform), dict):
        architecture = processor()
        for excluded_version in excluded_platforms_and_python_versions[sys.platform].get(architecture, list()):
            platform_string = f"{sys.platform}:{architecture}"

            if python_minor_version == version.parse(excluded_version).minor:
                raise ModuleNotFoundError(
                    f"\nThe package '{package_name}' is not available on the {platform_string} platform for "
                    f"Python version {excluded_version}!"
                )
    else:
        for excluded_version in excluded_platforms_and_python_versions.get(sys.platform, list()):
            if python_minor_version == version.parse(excluded_version).minor:
                raise ModuleNotFoundError(
                    f"\nThe package '{package_name}' is not available on the {sys.platform} platform for "
                    f"Python version {excluded_version}!"
                )

    if package_name in sys.modules:
        return sys.modules[package_name]

    if is_package_installed(package_name=package_name) is not None:
        return import_module(name=package_name)

    raise ModuleNotFoundError(
        f"\nThe required package'{package_name}' is not installed!\n"
        f"To install this package, please run\n\n\t{installation_instructions}\n"
    )


def get_format_summaries() -> dict[str, dict[str, Union[str, tuple[str, ...], None]]]:
    """Simple helper function for compiling high level summaries of all format interfaces and converters."""
    # Local scope import to avoid circularity
    from ..converters import converter_list
    from ..datainterfaces import interface_list

    summary_attribute_keys = ["display_name", "keywords", "associated_suffixes", "info"]

    combined_list = interface_list + converter_list
    summaries = {
        interface.__name__: {key: getattr(interface, key) for key in summary_attribute_keys}
        for interface in combined_list
    }

    return summaries



from pathlib import Path
from typing import Optional, Union

import numpy as np
from pydantic import FilePath, validate_call
from pynwb.file import NWBFile

from ....basetemporalalignmentinterface import BaseTemporalAlignmentInterface


class DeepLabCutInterface(BaseTemporalAlignmentInterface):
    """Data interface for DeepLabCut datasets."""

    display_name = "DeepLabCut"
    keywords = ("DLC",)
    associated_suffixes = (".h5",)
    info = "Interface for handling data from DeepLabCut."

    _timestamps = None

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the .h5 file output by dlc."
        source_schema["properties"]["config_file_path"]["description"] = "Path to .yml config file"
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        config_file_path: Optional[FilePath] = None,
        subject_name: str = "ind1",
        verbose: bool = True,
    ):
        """
        Interface for writing DLC's h5 files to nwb using dlc2nwb.

        Parameters
        ----------
        file_path : FilePath
            path to the h5 file output by dlc.
        config_file_path : FilePath, optional
            path to .yml config file
        subject_name : str, default: "ind1"
            the name of the subject for which the :py:class:`~pynwb.file.NWBFile` is to be created.
        verbose: bool, default: True
            controls verbosity.
        """
        from ._dlc_utils import _read_config

        file_path = Path(file_path)
        if "DLC" not in file_path.stem or ".h5" not in file_path.suffixes:
            raise IOError("The file passed in is not a DeepLabCut h5 data file.")

        self.config_dict = dict()
        if config_file_path is not None:
            self.config_dict = _read_config(config_file_path=config_file_path)
        self.subject_name = subject_name
        self.verbose = verbose
        super().__init__(file_path=file_path, config_file_path=config_file_path)

    def get_metadata(self):
        metadata = super().get_metadata()

        if self.config_dict:
            metadata["NWBFile"].update(
                session_description=self.config_dict["Task"],
                experimenter=[self.config_dict["scorer"]],
            )

        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        raise NotImplementedError(
            "Unable to retrieve the original unaltered timestamps for this interface! "
            "Define the `get_original_timestamps` method for this interface."
        )

    def get_timestamps(self) -> np.ndarray:
        raise NotImplementedError(
            "Unable to retrieve timestamps for this interface! Define the `get_timestamps` method for this interface."
        )

    def set_aligned_timestamps(self, aligned_timestamps: Union[list, np.ndarray]):
        """
        Set aligned timestamps vector for DLC data with user defined timestamps

        Parameters
        ----------
        aligned_timestamps : list, np.ndarray
            A timestamps vector.
        """
        self._timestamps = np.asarray(aligned_timestamps)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        container_name: str = "PoseEstimation",
    ):
        """
        Conversion from DLC output files to nwb. Derived from dlc2nwb library.

        Parameters
        ----------
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
        """
        from ._dlc_utils import add_subject_to_nwbfile

        add_subject_to_nwbfile(
            nwbfile=nwbfile,
            h5file=str(self.source_data["file_path"]),
            individual_name=self.subject_name,
            config_file=self.source_data["config_file_path"],
            timestamps=self._timestamps,
            pose_estimation_container_kwargs=dict(name=container_name),
        )



"""Base Pydantic models for the ZarrDatasetConfiguration."""

from typing import ClassVar, Literal, Type

import psutil
from hdmf_zarr import ZarrDataIO
from pydantic import Field

from ._base_backend import BackendConfiguration
from ._zarr_dataset_io import ZarrDatasetIOConfiguration


class ZarrBackendConfiguration(BackendConfiguration):
    """A model for matching collections of DatasetConfigurations specific to the Zarr backend."""

    backend: ClassVar[Literal["zarr"]] = "zarr"
    pretty_backend_name: ClassVar[Literal["Zarr"]] = "Zarr"
    data_io_class: ClassVar[Type[ZarrDataIO]] = ZarrDataIO

    dataset_configurations: dict[str, ZarrDatasetIOConfiguration] = Field(
        description=(
            "A mapping from object locations to their ZarrDatasetConfiguration specification that contains all "
            "information for writing the datasets to disk using the Zarr backend."
        )
    )
    number_of_jobs: int = Field(
        description=(
            "Number of jobs to use in parallel during write. Negative values, starting from -1, "
            "will use all the available CPUs (including logical), -2 is all except one, etc. "
            "This is equivalent to the pattern of indexing of "
            " `list(range(total_number_of_cpu))[number_of_jobs]`; for example, `-1` uses all available CPU, `-2` "
            "uses all except one, etc."
        ),
        ge=-psutil.cpu_count(),  # TODO: should we specify logical=False in cpu_count?
        le=psutil.cpu_count(),
        default=psutil.cpu_count() - 1,
    )



from __future__ import division  # for those using Python 2.6+

from collections import namedtuple
from datetime import datetime
from struct import calcsize, unpack


def _processheaders(curr_file, packet_fields):
    """
    :param curr_file:      {file} the current BR datafile to be processed
    :param packet_fields : {named tuple} the specific binary fields for the given header
    :return:               a fully unpacked and formatted tuple set of header information

    Read a packet from a binary data file and return a list of fields
    The amount and format of data read will be specified by the
    packet_fields container
    """

    # This is a lot in one line.  First I pull out all the format strings from
    # the basic_header_fields named tuple, then concatenate them into a string
    # with '<' at the front (for little endian format)
    packet_format_str = "<" + "".join([fmt for name, fmt, fun in packet_fields])

    # Calculate how many bytes to read based on the format strings of the header fields
    bytes_in_packet = calcsize(packet_format_str)
    packet_binary = curr_file.read(bytes_in_packet)

    # unpack the binary data from the header based on the format strings of each field.
    # This returns a list of data, but it's not always correctly formatted (eg, FileSpec
    # is read as ints 2 and 3, but I want it as '2.3'
    packet_unpacked = unpack(packet_format_str, packet_binary)

    # Create an iterator from the data list.  This allows a formatting function
    # to use more than one item from the list if needed, and the next formatting
    # function can pick up on the correct item in the list
    data_iter = iter(packet_unpacked)

    # create an empty dictionary from the name field of the packet_fields.
    # The loop below will fill in the values with formatted data by calling
    # each field's formatting function
    packet_formatted = dict.fromkeys([name for name, fmt, fun in packet_fields])
    for name, fmt, fun in packet_fields:
        packet_formatted[name] = fun(data_iter)
    curr_file.close()
    return packet_formatted


def _format_filespec(header_list):
    return str(next(header_list)) + "." + str(next(header_list))  # eg 2.3


def _format_timeorigin(header_list):
    year = next(header_list)
    month = next(header_list)
    _ = next(header_list)
    day = next(header_list)
    hour = next(header_list)
    minute = next(header_list)
    second = next(header_list)
    millisecond = next(header_list)
    return datetime(year, month, day, hour, minute, second, millisecond * 1000)


def _format_stripstring(header_list):
    string = bytes.decode(next(header_list), "latin-1")
    return string.split(STRING_TERMINUS, 1)[0]


def _format_none(header_list):
    return next(header_list)


FieldDef = namedtuple("FieldDef", ["name", "formatStr", "formatFnc"])
STRING_TERMINUS = "\x00"


def _parse_nsx_basic_header(nsx_file):
    nsx_basic_dict = [
        FieldDef("FileSpec", "2B", _format_filespec),  # 2 bytes   - 2 unsigned char
        FieldDef("BytesInHeader", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("Label", "16s", _format_stripstring),  # 16 bytes  - 16 char array
        FieldDef("Comment", "256s", _format_stripstring),  # 256 bytes - 256 char array
        FieldDef("Period", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("TimeStampResolution", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("TimeOrigin", "8H", _format_timeorigin),  # 16 bytes  - 8 uint16
        FieldDef("ChannelCount", "I", _format_none),
    ]  # 4 bytes   - uint32
    datafile = open(nsx_file, "rb")
    filetype_id = bytes.decode(datafile.read(8), "latin-1")
    if filetype_id == "NEURALSG":
        # this won't contain fields that can be added to NWBFile metadata
        return dict()
    return _processheaders(datafile, nsx_basic_dict)


def _parse_nev_basic_header(nev_file):
    nev_basic_dict = [
        FieldDef("FileTypeID", "8s", _format_stripstring),  # 8 bytes   - 8 char array
        FieldDef("FileSpec", "2B", _format_filespec),  # 2 bytes   - 2 unsigned char
        FieldDef("AddFlags", "H", _format_none),  # 2 bytes   - uint16
        FieldDef("BytesInHeader", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("BytesInDataPackets", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("TimeStampResolution", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("SampleTimeResolution", "I", _format_none),  # 4 bytes   - uint32
        FieldDef("TimeOrigin", "8H", _format_timeorigin),  # 16 bytes  - 8 x uint16
        FieldDef("CreatingApplication", "32s", _format_stripstring),  # 32 bytes  - 32 char array
        FieldDef("Comment", "256s", _format_stripstring),  # 256 bytes - 256 char array
        FieldDef("NumExtendedHeaders", "I", _format_none),
    ]
    datafile = open(nev_file, "rb")
    return _processheaders(datafile, nev_basic_dict)



import numpy as np
import pandas as pd
from pynwb.epoch import TimeIntervals


def convert_df_to_time_intervals(
    df: pd.DataFrame,
    table_name: str = "trials",
    table_description: str = "experimental trials",
    column_name_mapping: dict[str, str] = None,
    column_descriptions: dict[str, str] = None,
) -> TimeIntervals:
    """
    Convert a dataframe to a TimeIntervals object.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to convert.
    table_name : str, optional
        The name of the TimeIntervals object.
    table_description : str, optional
        The description of the TimeIntervals object.
    column_name_mapping: dict, optional
        If passed, rename subset of columns from key to value.
    column_descriptions: dict, optional
        Keys are the names of the columns (after renaming) and values are the descriptions. If not passed,
        the names of the columns are used as descriptions.

    Returns
    -------
    TimeIntervals

    """
    if column_name_mapping is not None:
        df.rename(columns=column_name_mapping, inplace=True)

    default_column_descriptions = dict(
        start_time="Start time of epoch, in seconds.",
        stop_time="Stop time of epoch, in seconds.",
    )

    if column_descriptions is None:
        column_descriptions = default_column_descriptions
    else:
        column_descriptions = dict(default_column_descriptions, **column_descriptions)

    time_intervals = TimeIntervals(name=table_name, description=table_description)
    if "start_time" not in df:
        raise ValueError(f"df must contain a column named 'start_time'. Existing columns: {df.columns.to_list()}")
    if "stop_time" not in df:
        df["stop_time"] = np.r_[df["start_time"][1:].to_numpy(), np.nan]
    for col in df:
        if col not in ("start_time", "stop_time"):
            time_intervals.add_column(col, column_descriptions.get(col, col))
    for i, row in df.iterrows():
        time_intervals.add_row(row.to_dict())

    return time_intervals



from pathlib import Path

from pydantic import FilePath, validate_call

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ..basesortingextractorinterface import BaseSortingExtractorInterface
from ....utils import DeepDict


class PlexonRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface class for converting Plexon data.

    Uses the :py:class:`~spikeinterface.extractors.PlexonRecordingExtractor`.
    """

    display_name = "Plexon Recording"
    associated_suffixes = (".plx",)
    info = "Interface for Plexon recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the .plx file."
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        verbose: bool = True,
        es_key: str = "ElectricalSeries",
        stream_name: str = "WB-Wideband",
    ):
        """
        Load and prepare data for Plexon.

        Parameters
        ----------
        file_path : str or Path
            Path to the .plx file.
        verbose : bool, default: True
            Allows verbosity.
        es_key : str, default: "ElectricalSeries"
        stream_name: str, optional
            Only pass a stream if you modified the channel prefixes in the Plexon file and you know the prefix of
            the wideband data.
        """

        invalid_stream_names = ["FPl-Low Pass Filtered", "SPKC-High Pass Filtered", "AI-Auxiliary Input"]
        assert stream_name not in invalid_stream_names, f"Invalid stream name: {stream_name}"

        super().__init__(file_path=file_path, verbose=verbose, es_key=es_key, stream_name=stream_name)

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()
        neo_reader = self.recording_extractor.neo_reader

        if hasattr(neo_reader, "raw_annotations"):
            block_ind = self.recording_extractor.block_index
            neo_metadata = neo_reader.raw_annotations["blocks"][block_ind]
            if "rec_datetime" in neo_metadata:
                metadata["NWBFile"].update(session_start_time=neo_metadata["rec_datetime"])

        return metadata


class Plexon2RecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface class for converting Plexon2 data.

    Uses the :py:class:`~spikeinterface.extractors.Plexon2RecordingExtractor`.
    """

    display_name = "Plexon2 Recording"
    associated_suffixes = (".pl2",)
    info = "Interface for Plexon2 recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the .pl2 file."
        return source_schema

    @validate_call
    def __init__(self, file_path: FilePath, verbose: bool = True, es_key: str = "ElectricalSeries"):
        """
        Load and prepare data for Plexon.

        Parameters
        ----------
        file_path : str or Path
            Path to the .plx file.
        verbose : bool, default: True
            Allows verbosity.
        es_key : str, default: "ElectricalSeries"
        """
        # TODO: when neo version 0.14.4 is out or higher change this to stream_name for clarify
        import neo
        from packaging.version import Version

        neo_version = Version(neo.__version__)
        if neo_version <= Version("0.13.3"):
            stream_id = "3"
        else:
            stream_id = "WB"
        assert Path(file_path).is_file(), f"Plexon file not found in: {file_path}"
        super().__init__(
            file_path=file_path,
            verbose=verbose,
            es_key=es_key,
            stream_id=stream_id,
            all_annotations=True,
        )

    def get_metadata(self) -> DeepDict:
        metadata = super().get_metadata()

        neo_reader = self.recording_extractor.neo_reader

        block_ind = self.recording_extractor.block_index
        neo_metadata = neo_reader.raw_annotations["blocks"][block_ind]
        metadata["NWBFile"].update(session_start_time=neo_metadata["m_CreatorDateTime"])

        return metadata


class PlexonSortingInterface(BaseSortingExtractorInterface):
    """
    Primary data interface class for converting Plexon spiking data.

    Uses :py:class:`~spikeinterface.extractors.PlexonSortingExtractor`.
    """

    display_name = "Plexon Sorting"
    associated_suffixes = (".plx",)
    info = "Interface for Plexon sorting data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the plexon spiking data (.plx file)."
        return source_schema

    @validate_call
    def __init__(self, file_path: FilePath, verbose: bool = True):
        """
        Load and prepare data for Plexon.

        Parameters
        ----------
        file_path: FilePathType
            Path to the plexon spiking data (.plx file).
        verbose: bool, default: True
            Allows verbosity.
        """
        super().__init__(file_path=file_path, verbose=verbose)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        neo_reader = self.sorting_extractor.neo_reader

        if hasattr(neo_reader, "raw_annotations"):
            block_ind = self.sorting_extractor.block_index
            neo_metadata = neo_reader.raw_annotations["blocks"][block_ind]

            if "rec_datetime" in neo_metadata:
                metadata["NWBFile"].update(session_start_time=neo_metadata["rec_datetime"])

        return metadata



from typing import Optional, Union

import numpy as np

ArrayType = Union[list, np.ndarray]
OptionalArrayType = Optional[ArrayType]
FloatType = float
IntType = Union[int, np.integer]


# TODO: remove after 3/1/2025
def __getattr__(name):
    from typing import Optional
    from warnings import warn

    from pydantic import DirectoryPath, FilePath

    if name == "FilePathType":
        message = (
            "The 'FilePathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `pydantic.FilePath` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return FilePath
    if name == "OptionalFilePathType":
        message = (
            "The 'OptionalFilePathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `typing.Optional[pydantic.FilePath]` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return Optional[FilePath]
    if name == "FolderPathType":
        message = (
            "The 'FolderPathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `pydantic.DirectoryPath` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return DirectoryPath
    if name == "OptionalFolderPathType":
        message = (
            "The 'OptionalFolderPathType' type has been deprecated and will be removed after 3/1/2025. "
            "Please use `typing.Optional[pydantic.DirectoryPath]` instead."
        )
        warn(message=message, category=DeprecationWarning, stacklevel=2)

        return Optional[DirectoryPath]



from pydantic import FilePath

from ..basesegmentationextractorinterface import BaseSegmentationExtractorInterface


class CnmfeSegmentationInterface(BaseSegmentationExtractorInterface):
    """Data interface for constrained non-negative matrix factorization (CNMFE) segmentation extractor."""

    display_name = "CNMFE Segmentation"
    associated_suffixes = (".mat",)
    info = "Interface for constrained non-negative matrix factorization (CNMFE) segmentation."

    def __init__(self, file_path: FilePath, verbose: bool = True):
        super().__init__(file_path=file_path)
        self.verbose = verbose



from datetime import datetime
from pathlib import Path
from typing import Union
from warnings import warn

from lxml import etree
from pydantic import DirectoryPath


def _get_session_start_time(element: etree.Element) -> Union[datetime, None]:
    """
    Get the session start time from the settings.xml file.
    Returns the session start time as a datetime object.
    When the session start time is not found in the XML, a warning is issued and None is returned.

    Parameters
    ----------
    element : etree.Element
        The root element of the XML tree.
    """
    date_str = element.findtext("./INFO/DATE")
    if date_str is None:
        warn(message="Could not fetch session start time from settings.xml file.", category=UserWarning)
        return None

    session_start_time = datetime.strptime(date_str, "%d %b %Y %H:%M:%S")
    return session_start_time


def _read_settings_xml(folder_path: DirectoryPath) -> etree.Element:
    """
    Read the settings.xml file from an OpenEphys binary recording folder.
    Returns the root element of the XML tree.

    Parameters
    ----------
    folder_path : FolderPathType
        The path to the folder containing the settings.xml file.
    """
    folder_path = Path(folder_path)

    xml_file_paths = list(folder_path.rglob("settings.xml"))
    if not len(xml_file_paths) == 1:
        raise ValueError(
            f"Unable to identify the OpenEphys folder structure! "
            f"Please check that your `folder_path` contains a settings.xml file and sub-folders of the "
            "following form: 'experiment<index>' -> 'recording<index>' -> 'continuous'."
        )
    xml_file_path = xml_file_paths[0]
    tree = etree.parse(str(xml_file_path))
    return tree.getroot()



import json
import re
import warnings
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Union

import numpy as np
from pydantic import FilePath, validate_call
from pynwb.behavior import Position, SpatialSeries
from pynwb.file import NWBFile

from ....basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from ....tools import get_module
from ....utils import calculate_regular_series_rate


class FicTracDataInterface(BaseTemporalAlignmentInterface):
    """Data interface for FicTrac datasets."""

    display_name = "FicTrac"
    keywords = (
        "fictrack",
        "visual tracking",
        "fictive path",
        "spherical treadmill",
        "visual fixation",
    )
    associated_suffixes = (".dat",)
    info = "Interface for FicTrac .dat files."

    timestamps_column = 21
    # Columns in the .dat binary file with the data. The full description of the header can be found in:
    # https://github.com/rjdmoore/fictrac/blob/master/doc/data_header.txt
    columns_in_dat_file = [
        "frame_counter",
        "rotation_delta_x_cam",
        "rotation_delta_y_cam",
        "rotation_delta_z_cam",
        "rotation_delta_error",
        "rotation_delta_x_lab",
        "rotation_delta_y_lab",
        "rotation_delta_z_lab",
        "rotation_x_cam",
        "rotation_y_cam",
        "rotation_z_cam",
        "rotation_x_lab",
        "rotation_y_lab",
        "rotation_z_lab",
        "x_pos_radians_lab",
        "y_pos_radians_lab",
        "animal_heading",
        "movement_direction",
        "movement_speed",
        "forward_motion_lab",
        "side_motion_lab",
        "timestamp",
        "sequence_counter",
        "delta_timestamp",
        "alt_timestamp",
    ]

    column_to_nwb_mapping = spatial_series_descriptions = {
        "rotation_delta_cam": {
            "column_in_dat_file": ["rotation_delta_x_cam", "rotation_delta_y_cam", "rotation_delta_z_cam"],
            "spatial_series_name": "SpatialSeriesRotationDeltaCameraFrame",
            "description": (
                "Change in orientation since last frame from the camera's perspective. "
                "x: rotation to the sphere's right (pitch), "
                "y: rotation under the sphere (yaw), "
                "z: rotation behind the sphere (roll)"
            ),
            "reference_frame": "camera",
        },
        "rotation_delta_lab": {
            "column_in_dat_file": ["rotation_delta_x_lab", "rotation_delta_y_lab", "rotation_delta_z_lab"],
            "spatial_series_name": "SpatialSeriesRotationDeltaLabFrame",
            "description": (
                "Change in orientation since last frame from the lab's perspective. "
                "x: rotation in front of subject (roll), "
                "y: rotation to subject's right (pitch), "
                "z: rotation under the subject (yaw)"
            ),
            "reference_frame": "lab",
        },
        "rotation_cam": {
            "column_in_dat_file": ["rotation_x_cam", "rotation_y_cam", "rotation_z_cam"],
            "spatial_series_name": "SpatialSeriesRotationCameraFrame",
            "description": (
                "Orientation in radians from the camera's perspective. "
                "x: rotation to the sphere's right (pitch), "
                "y: rotation under the sphere (yaw), "
                "z: rotation behind the sphere (roll)"
            ),
            "reference_frame": "camera",
        },
        "rotation_lab": {
            "column_in_dat_file": ["rotation_x_lab", "rotation_y_lab", "rotation_z_lab"],
            "spatial_series_name": "SpatialSeriesRotationLabFrame",
            "description": (
                "Orientation in radians from the lab's perspective. "
                "x: rotation in front of subject (roll), "
                "y: rotation to subject's right (pitch), "
                "z: rotation under the subject (yaw)"
            ),
            "reference_frame": "lab",
        },
        "animal_heading": {
            "column_in_dat_file": ["animal_heading"],
            "spatial_series_name": "SpatialSeriesAnimalHeading",
            "description": "Animal's heading direction in radians from the lab's perspective.",
            "reference_frame": "lab",
        },
        "movement_direction": {
            "column_in_dat_file": ["movement_direction"],
            "spatial_series_name": "SpatialSeriesMovementDirection",
            "description": (
                "Instantaneous running direction of the animal in the lab coordinates. "
                "Direction inferred by the ball's rotation (roll and pitch)"
            ),
            "reference_frame": "lab",
        },
        "movement_speed": {
            "column_in_dat_file": ["movement_speed"],
            "spatial_series_name": "SpatialSeriesMovementSpeed",
            "description": "Instantaneous running speed of the animal in radians per frame.",
            "reference_frame": "lab",
        },
        "position_lab": {
            "column_in_dat_file": ["x_pos_radians_lab", "y_pos_radians_lab"],
            "spatial_series_name": "SpatialSeriesPosition",
            "description": (
                "x and y positions in the lab frame in radians, inferred by integrating " "the rotation over time."
            ),
            "reference_frame": "lab",
        },
        "integrated_motion": {
            "column_in_dat_file": ["forward_motion_lab", "side_motion_lab"],
            "spatial_series_name": "SpatialSeriesIntegratedMotion",
            "description": ("Integrated x/y position of the sphere in laboratory coordinates, neglecting heading."),
            "reference_frame": "lab",
        },
        "rotation_delta_error": {
            "spatial_series_name": "SpatialSeriesRotationDeltaError",
            "column_in_dat_file": ["rotation_delta_error"],
            "description": "Error in rotation delta in radians from the lab's perspective.",
            "reference_frame": "lab",
        },
    }

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to the .dat file (the output of fictrac)"
        return source_schema

    @validate_call
    def __init__(
        self,
        file_path: FilePath,
        radius: Optional[float] = None,
        configuration_file_path: Optional[FilePath] = None,
        verbose: bool = True,
    ):
        """
        Interface for writing FicTrac files to nwb.

        Parameters
        ----------
        file_path : FilePath
            Path to the .dat file (the output of fictrac)
        radius : float, optional
            The radius of the ball in meters. If provided the radius is stored as a conversion factor
            and the units are set to meters. If not provided the units are set to radians.
        configuration_file_path : FilePath, optional
            Path to the .txt file with the configuration metadata. Usually called config.txt
        verbose : bool, default: True
            controls verbosity. ``True`` by default.
        """
        self.file_path = Path(file_path)
        self.verbose = verbose
        self.radius = radius
        self.configuration_file_path = None
        super().__init__(file_path=file_path)

        self.configuration_file_path = configuration_file_path
        if self.configuration_file_path is None and (self.file_path.parent / "config.txt").is_file():
            self.configuration_file_path = self.file_path.parent / "config.txt"

        self.configuration_metadata = None
        if self.configuration_file_path is not None:
            self.configuration_metadata = parse_fictrac_config(file_path=self.configuration_file_path)

        self._timestamps = None
        self._starting_time = None

    def get_metadata(self):
        metadata = super().get_metadata()

        session_start_time = extract_session_start_time(
            file_path=self.file_path,
            configuration_file_path=self.configuration_file_path,
        )
        if session_start_time:
            metadata["NWBFile"].update(session_start_time=session_start_time)

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        compression: Optional[str] = None,  # TODO: remove completely after 10/1/2024
        compression_opts: Optional[int] = None,  # TODO: remove completely after 10/1/2024
    ):
        """
        Parameters
        ----------
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict, optional
            metadata info for constructing the nwb file.
        """
        import pandas as pd

        # TODO: remove completely after 10/1/2024
        if compression is not None or compression_opts is not None:
            warnings.warn(
                message=(
                    "Specifying compression methods and their options at the level of tool functions has been deprecated. "
                    "Please use the `configure_backend` tool function for this purpose."
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )

        fictrac_data_df = pd.read_csv(self.file_path, sep=",", header=None, names=self.columns_in_dat_file)

        # Get the timestamps
        timestamps = self.get_timestamps()
        starting_time = timestamps[0]

        # Note: Returns the timestamps from the sources in FicTrac (.e.g OpenCV videos, PGR or Basler). The most common  is OpenCV with all the caveats that come with it.
        # video timestamp extracted with OpenCV with all the caveats that come with it.
        rate = calculate_regular_series_rate(series=timestamps)  # Returns None if the series is not regular
        if rate:
            write_timestamps = False
        else:
            write_timestamps = True

        position_container = Position(name="FicTrac")
        # TODO: make FicTrac extension of the Position container to attach these specific attributes with documentation
        if self.configuration_metadata is not None:
            comments = json.dumps(self.configuration_metadata)

        for data_dict in self.column_to_nwb_mapping.values():
            spatial_series_kwargs = dict(
                name=data_dict["spatial_series_name"],
                description=data_dict["description"],
                reference_frame=data_dict["reference_frame"],
            )

            if self.configuration_metadata is not None:
                spatial_series_kwargs["comments"] = comments

            column_in_dat_file = data_dict["column_in_dat_file"]
            data = fictrac_data_df[column_in_dat_file].to_numpy()
            if self.radius is not None:
                spatial_series_kwargs["conversion"] = self.radius
                units = "meters"
            else:
                units = "radians"

            spatial_series_kwargs.update(data=data, unit=units)

            if write_timestamps:
                spatial_series_kwargs["timestamps"] = timestamps
            else:
                spatial_series_kwargs["rate"] = rate
                spatial_series_kwargs["starting_time"] = starting_time

            # Create the spatial series and add it to the container
            spatial_series = SpatialSeries(**spatial_series_kwargs)
            position_container.add_spatial_series(spatial_series)

        # Add the container to the processing module
        processing_module = get_module(nwbfile=nwbfile, name="behavior")
        processing_module.add_data_interface(position_container)

    def get_original_timestamps(self):
        """
        Retrieve and correct timestamps from a FicTrac data file.

        This function addresses two specific issues with timestamps in FicTrac data:

        1. Resetting Initial Timestamp
           In some instances, FicTrac replaces the initial timestamp (0) with the system time. This commonly occurs
           when the data source is a video file, and OpenCV reports the first timestamp as 0. Since OpenCV also
           uses 0 as a marker for invalid values, FicTrac defaults to system time in that case. This leads to
           inconsistent timestamps like [system_time, t1, t2, t3, ...]. The function corrects this by resetting the
           first timestamp back to 0 when a negative difference is detected between the first two timestamps.
        2. Re-centering Unix Epoch Time
           If timestamps are in Unix epoch time format (time since 1970-01-01 00:00:00 UTC), this function re-centers
           the time series by subtracting the first timestamp. This adjustment ensures that timestamps represent the
           elapsed time since the start of the experiment rather than the Unix epoch. This case appears when one of the
           sources of data in FicTrac (such as PGR or Basler) lacks a timestamp extraction method. FicTrac
           then falls back to using the system time, which is in Unix epoch format.

        Returns
        -------
        np.ndarray
            An array of corrected timestamps, in seconds.

        Notes
        -----
        - The issue of the initial timestamp replacement appears in FicTrac 2.1.1 and earlier versions.
        - Re-centering is essential for timestamps in Unix epoch format as timestamps in an NWB file must be relative
        to the start of the session. The heuristic here is to check if the first timestamp is larger than the length
        of a 10-year experiment in seconds. If so, it's assumed that the timestamps are in Unix epoch format.

        References
        ----------
        Issue discussion on FicTrac's timestamp inconsistencies:
        https://github.com/rjdmoore/fictrac/issues/29
        """

        import pandas as pd

        fictrac_data_df = pd.read_csv(self.file_path, sep=",", header=None, usecols=[self.timestamps_column])

        timestamps = fictrac_data_df[self.timestamps_column].values / 1000.0  # Transform to seconds

        # Correct for the case when only the first timestamp was replaced by system time
        first_difference = timestamps[1] - timestamps[0]
        if first_difference < 0:
            timestamps[0] = 0.0

        # Heuristic to test if timestamps are in Unix epoch
        length_in_seconds_of_a_10_year_experiment = 10 * 365 * 24 * 60 * 60
        all_timestamps_are_in_unix_epoch = np.all(timestamps > length_in_seconds_of_a_10_year_experiment)
        if all_timestamps_are_in_unix_epoch:
            timestamps = timestamps - timestamps[0]
        # TODO: If we agree to ALWAYS constrain timestamps to be relative to the start of the session, we can
        # Always shift here and remove the heuristic above.

        return timestamps

    def get_timestamps(self):
        timestamps = self._timestamps if self._timestamps is not None else self.get_original_timestamps()
        if self._starting_time is not None:
            timestamps = timestamps + self._starting_time

        return timestamps

    def set_aligned_timestamps(self, aligned_timestamps):
        self._timestamps = aligned_timestamps

    def set_aligned_starting_time(self, aligned_starting_time):
        self._starting_time = aligned_starting_time


def extract_session_start_time(
    file_path: FilePath,
    configuration_file_path: Optional[FilePath] = None,
) -> Union[datetime, None]:
    """
    Extract the session start time from a FicTrac data file or its configuration file.

    The session start time is determined from the data file if the timestamps are in Unix epoch format. If not, the
    function defaults to extracting the date from the configuration file and assuming that the start time is midnight.
    If neither of these methods works, the function returns None.

    The session start time, has two different meanings depending on the source of the FicTrac data:
    - For video file sources (.avi, .mp4, etc.), the session start time corresponds to the time when the
    FicTrac analysis commenced. That is, the session start time reflects the analysis time rather than
    the actual start of the experiment.
    - For camera sources (such as PGR or Basler), the session start time is either the time reported by the camera
    or the system time if the camera's SDK does not provide timestamps to Fictrac. In both cases, this time is
    the experiment start time, barring synchronization issues.

    Parameters
    ----------
    file_path : FilePath
        Path to the FicTrac data file.
    configuration_file_path : FilePath, optional
        Path to the FicTrac configuration file. If omitted, the function defaults to searching for
        "fictrac_config.txt" in the same directory as the data file.

    Returns
    -------
    datetime | None
        The session start time of in UTC as a datetime object. `None` if the session start time cannot be extracted.

    """
    with open(file_path, "r") as file:
        first_line = file.readline()

    timestamps_column = FicTracDataInterface.timestamps_column
    first_timestamp = float(first_line.split(",")[timestamps_column]) / 1000.0  # Convert to seconds

    # Heuristic to test if timestamps are in Unix epoch
    length_in_seconds_of_a_10_year_experiment = 10 * 365 * 24 * 60 * 60
    if first_timestamp > length_in_seconds_of_a_10_year_experiment:
        utc_timestamp = first_timestamp
        return datetime.utcfromtimestamp(utc_timestamp).replace(tzinfo=timezone.utc)

    if configuration_file_path is None:
        configuration_file_path = file_path.parent / "fictrac_config.txt"
    if configuration_file_path.is_file():
        configuration_file = parse_fictrac_config(configuration_file_path)
        session_start_time = datetime.strptime(configuration_file.get("build_date", ""), "%b %d %Y")
        # Set the time to midnight UTC from the extracted date
        return session_start_time.replace(hour=0, minute=0, second=0, microsecond=0, tzinfo=timezone.utc)

    return None


def parse_fictrac_config(file_path: FilePath) -> dict:
    """
    Parse a FicTrac configuration file and return a dictionary of its parameters.

    Parameters
    ----------
    file_path : FilePath
        Path to the configuration file in txt format.

    Returns
    -------
    dict
        A dictionary where the keys are the parameter names and the values are the parameter values.

    """

    def parse_bool(x):
        return x.lower() == "y"

    def parse_vec_int(x):
        return [int(val) for val in x.replace("{", "").replace("}", "").split(",")]

    def parse_vec_float(x):
        return [float(val) for val in x.replace("{", "").replace("}", "").split(",")]

    def parse_roi_ignr(x):
        innner_vectors = x.strip("{}").strip().split("}, {")
        return [parse_vec_int(group) for group in innner_vectors]

    def parse_int_or_string(x):
        value = int(x) if x.isdigit() else x
        return value

    key_parsers = {
        "src_fn": parse_int_or_string,
        "vfov": float,
        "do_display": parse_bool,
        "save_debug": parse_bool,
        "save_raw": parse_bool,
        "sock_port": int,
        "com_baud": int,
        "fisheye": parse_bool,
        "q_factor": int,
        "src_fps": float,
        "max_bad_frames": int,
        "opt_do_global": parse_bool,
        "opt_max_err": float,
        "thr_ratio": float,
        "thr_win_pc": float,
        "opt_max_evals": int,
        "opt_bound": float,
        "opt_tol": float,
        "c2a_cnrs_xy": parse_vec_int,
        "c2a_cnrs_yz": parse_vec_int,
        "c2a_cnrs_xz": parse_vec_int,
        "c2a_r": parse_vec_float,
        "c2a_t": parse_vec_float,
        "roi_circ": parse_vec_int,
        "roi_c": parse_vec_float,
        "roi_r": float,
        "roi_ignr": parse_roi_ignr,
    }

    # Open and read the file
    with open(file_path, "r") as f:
        file_lines = f.readlines()

    parsed_config = {}

    # Parse header line
    header_line = file_lines[0]

    version, build_date = re.search(
        r"FicTrac (v[0-9.]+) config file \(build date ([A-Za-z0-9 ]+)\)", header_line
    ).groups()
    parsed_config["version"] = version
    parsed_config["build_date"] = build_date

    # Parse the configuration lines
    configuration_lines = (line for line in file_lines[1:] if ":" in line)
    for line in configuration_lines:
        key, value = line.split(":", 1)
        key = key.strip()
        value = value.strip()

        parser = key_parsers.get(key, lambda x: x)
        parsed_config[key] = parser(value)
    return parsed_config



"""Base Pydantic models for DatasetInfo and DatasetConfiguration."""

import math
from abc import ABC, abstractmethod
from typing import Any, Literal, Union

import h5py
import numcodecs
import numpy as np
import zarr
from hdmf import Container
from hdmf.data_utils import GenericDataChunkIterator
from hdmf.utils import get_data_shape
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    InstanceOf,
    PositiveInt,
    model_validator,
)
from pynwb import NWBFile
from typing_extensions import Self

from neuroconv.utils.str_utils import human_readable_size

from ._pydantic_pure_json_schema_generator import PureJSONSchemaGenerator
from ...hdmf import SliceableDataChunkIterator


def _recursively_find_location_in_memory_nwbfile(current_location: str, neurodata_object: Container) -> str:
    """
    Method for determining the location of a neurodata object within an in-memory NWBFile object.

    Distinct from methods from other packages, such as the NWB Inspector, which rely on such files being read from disk.
    """
    parent = neurodata_object.parent
    if isinstance(parent, NWBFile):
        # Items in defined top-level places like acquisition, intervals, etc. do not act as 'containers'
        # in that they do not set the `.parent` attribute; ask if object is in their in-memory dictionaries instead
        for parent_field_name, parent_field_value in parent.fields.items():
            if isinstance(parent_field_value, dict) and neurodata_object.name in parent_field_value:
                return parent_field_name + "/" + neurodata_object.name + "/" + current_location
        return neurodata_object.name + "/" + current_location
    return _recursively_find_location_in_memory_nwbfile(
        current_location=neurodata_object.name + "/" + current_location, neurodata_object=parent
    )


def _find_location_in_memory_nwbfile(neurodata_object: Container, field_name: str) -> str:
    """
    More readable call for the recursive location finder for a field of a neurodata object in an in-memory NWBFile.

    The recursive method forms from the buttom-up using the initial 'current_location' of the field itself.
    """
    return _recursively_find_location_in_memory_nwbfile(current_location=field_name, neurodata_object=neurodata_object)


def _infer_dtype_of_list(list_: list[Union[int, float, list]]) -> np.dtype:
    """
    Attempt to infer the dtype of values in an arbitrarily sized and nested list.

    Relies on the ability of the numpy.array call to cast the list as an array so the 'dtype' attribute can be used.
    """
    for item in list_:
        if isinstance(item, list):
            dtype = _infer_dtype_of_list(list_=item)
            if dtype is not None:
                return dtype
        else:
            return np.array([item]).dtype

    raise ValueError("Unable to determine the dtype of values in the list.")


def _infer_dtype(dataset: Union[h5py.Dataset, zarr.Array]) -> np.dtype:
    """Attempt to infer the dtype of the contained values of the dataset."""
    if hasattr(dataset, "dtype"):
        data_type = np.dtype(dataset.dtype)
        return data_type

    if isinstance(dataset, list):
        return _infer_dtype_of_list(list_=dataset)

    # Think more on if there is a better way to handle this fallback
    data_type = np.dtype("object")
    return data_type


class DatasetIOConfiguration(BaseModel, ABC):
    """A data model for configuring options about an object that will become a HDF5 or Zarr Dataset in the file."""

    model_config = ConfigDict(validate_assignment=True)  # Re-validate model on mutation

    # Immutable fields about the dataset
    object_id: str = Field(description="The UUID of the neurodata object containing the dataset.", frozen=True)
    location_in_file: str = Field(
        description=(
            "The location of the this dataset within the in-memory NWBFile relative to the top-level root, "
            "e.g. 'acquisition/ElectricalSeries/data'."
        ),
        frozen=True,
    )
    dataset_name: Literal["data", "timestamps"] = Field(description="The reference name of the dataset.", frozen=True)
    dtype: InstanceOf[np.dtype] = Field(description="The data type of elements of this dataset.", frozen=True)
    full_shape: tuple[int, ...] = Field(description="The maximum shape of the entire dataset.", frozen=True)

    # User specifiable fields
    chunk_shape: Union[tuple[PositiveInt, ...], None] = Field(
        description=(
            "The specified shape to use when chunking the dataset. "
            "For optimized streaming speeds, a total size of around 10 MB is recommended."
        ),
    )
    buffer_shape: Union[tuple[int, ...], None] = Field(
        description=(
            "The specified shape to use when iteratively loading data into memory while writing the dataset. "
            "For optimized writing speeds and minimal RAM usage, a total size of around 1 GB is recommended."
        ),
    )
    compression_method: Union[
        str, InstanceOf[h5py._hl.filters.FilterRefBase], InstanceOf[numcodecs.abc.Codec], None
    ] = Field(
        description="The specified compression method to apply to this dataset. Set to `None` to disable compression.",
    )
    compression_options: Union[dict[str, Any], None] = Field(
        default=None, description="The optional parameters to use for the specified compression method."
    )

    @abstractmethod
    def get_data_io_kwargs(self) -> dict[str, Any]:
        """
        Fetch the properly structured dictionary of input arguments.

        Should be passed directly as dynamic keyword arguments (**kwargs) into a H5DataIO or ZarrDataIO.
        """
        raise NotImplementedError

    def __str__(self) -> str:
        """
        Not overriding __repr__ as this is intended to render only when wrapped in print().

        Reason being two-fold; a standard `repr` is intended to be slightly more machine-readable / a more basic
        representation of the true object state. But then also because an iterable of these objects, such as a
        `list[DatasetConfiguration]`, would print out the nested representations, which only look good when using the
        basic `repr` (that is, this fancy string print-out does not look good when nested in another container).
        """
        size_in_bytes = math.prod(self.full_shape) * self.dtype.itemsize
        maximum_ram_usage_per_iteration_in_bytes = math.prod(self.buffer_shape) * self.dtype.itemsize
        disk_space_usage_per_chunk_in_bytes = math.prod(self.chunk_shape) * self.dtype.itemsize

        string = (
            f"\n{self.location_in_file}"
            f"\n{'-' * len(self.location_in_file)}"
            f"\n  dtype : {self.dtype}"
            f"\n  full shape of source array : {self.full_shape}"
            f"\n  full size of source array : {human_readable_size(size_in_bytes)}"
            "\n"
            f"\n  buffer shape : {self.buffer_shape}"
            f"\n  expected RAM usage : {human_readable_size(maximum_ram_usage_per_iteration_in_bytes)}"
            "\n"
            f"\n  chunk shape : {self.chunk_shape}"
            f"\n  disk space usage per chunk : {human_readable_size(disk_space_usage_per_chunk_in_bytes)}"
            "\n"
        )
        if self.compression_method is not None:
            string += f"\n  compression method : {self.compression_method}"
        if self.compression_options is not None:
            string += f"\n  compression options : {self.compression_options}"
        if self.compression_method is not None or self.compression_options is not None:
            string += "\n"
        # TODO: would be cool to include estimate of ratio too (determined via stub file perhaps?)

        return string

    @model_validator(mode="before")
    def validate_all_shapes(cls, values: dict[str, Any]) -> dict[str, Any]:
        location_in_file = values["location_in_file"]
        dataset_name = values["dataset_name"]

        assert (
            dataset_name == location_in_file.split("/")[-1]
        ), f"The `dataset_name` ({dataset_name}) does not match the end of the `location_in_file` ({location_in_file})!"

        chunk_shape = values["chunk_shape"]
        buffer_shape = values["buffer_shape"]
        full_shape = values["full_shape"]

        if len(chunk_shape) != len(buffer_shape):
            raise ValueError(
                f"{len(chunk_shape)=} does not match {len(buffer_shape)=} for dataset at location '{location_in_file}'!"
            )
        if len(buffer_shape) != len(full_shape):
            raise ValueError(
                f"{len(buffer_shape)=} does not match {len(full_shape)=} for dataset at location '{location_in_file}'!"
            )

        if any(chunk_axis <= 0 for chunk_axis in chunk_shape):
            raise ValueError(
                f"Some dimensions of the {chunk_shape=} are less than or equal to zero for dataset at "
                f"location '{location_in_file}'!"
            )
        if any(buffer_axis <= 0 for buffer_axis in buffer_shape):
            raise ValueError(
                f"Some dimensions of the {buffer_shape=} are less than or equal to zero for dataset at "
                f"location '{location_in_file}'!"
            )

        if any(chunk_axis > buffer_axis for chunk_axis, buffer_axis in zip(chunk_shape, buffer_shape)):
            raise ValueError(
                f"Some dimensions of the {chunk_shape=} exceed the {buffer_shape=} for dataset at "
                f"location '{location_in_file}'!"
            )
        if any(buffer_axis > full_axis for buffer_axis, full_axis in zip(buffer_shape, full_shape)):
            raise ValueError(
                f"Some dimensions of the {buffer_shape=} exceed the {full_shape=} for dataset at "
                f"location '{location_in_file}'!"
            )

        if any(
            buffer_axis % chunk_axis != 0
            for chunk_axis, buffer_axis, full_axis in zip(chunk_shape, buffer_shape, full_shape)
            if buffer_axis != full_axis
        ):
            raise ValueError(
                f"Some dimensions of the {chunk_shape=} do not evenly divide the {buffer_shape=} for dataset at "
                f"location '{location_in_file}'!"
            )

        return values

    # Pydantic models have several API calls for retrieving the schema - override all of them to work
    @classmethod
    def schema(cls, **kwargs) -> dict[str, Any]:
        return cls.model_json_schema(**kwargs)

    @classmethod
    def schema_json(cls, **kwargs) -> dict[str, Any]:
        return cls.model_json_schema(**kwargs)

    @classmethod
    def model_json_schema(cls, **kwargs) -> dict[str, Any]:
        assert "mode" not in kwargs, "The 'mode' of this method is fixed to be 'validation' and cannot be changed."
        assert "schema_generator" not in kwargs, "The 'schema_generator' of this method cannot be changed."
        return super().model_json_schema(mode="validation", schema_generator=PureJSONSchemaGenerator, **kwargs)

    @classmethod
    def from_neurodata_object(cls, neurodata_object: Container, dataset_name: Literal["data", "timestamps"]) -> Self:
        """
        Construct an instance of a DatasetIOConfiguration for a dataset in a neurodata object in an NWBFile.

        Parameters
        ----------
        neurodata_object : hdmf.Container
            The neurodata object containing the field that will become a dataset when written to disk.
        dataset_name : "data" or "timestamps"
            The name of the field that will become a dataset when written to disk.
            Some neurodata objects can have multiple such fields, such as `pynwb.TimeSeries` which can have both `data`
            and `timestamps`, each of which can be configured separately.
        """
        location_in_file = _find_location_in_memory_nwbfile(neurodata_object=neurodata_object, field_name=dataset_name)

        candidate_dataset = getattr(neurodata_object, dataset_name)
        full_shape = get_data_shape(data=candidate_dataset)
        dtype = _infer_dtype(dataset=candidate_dataset)

        if isinstance(candidate_dataset, GenericDataChunkIterator):
            chunk_shape = candidate_dataset.chunk_shape
            buffer_shape = candidate_dataset.buffer_shape
            compression_method = "gzip"
        elif dtype != np.dtype("object"):
            chunk_shape = SliceableDataChunkIterator.estimate_default_chunk_shape(
                chunk_mb=10.0, maxshape=full_shape, dtype=np.dtype(dtype)
            )
            buffer_shape = SliceableDataChunkIterator.estimate_default_buffer_shape(
                buffer_gb=0.5, chunk_shape=chunk_shape, maxshape=full_shape, dtype=np.dtype(dtype)
            )
            compression_method = "gzip"
        elif dtype == np.dtype("object"):  # Unclear what default chunking/compression should be for compound objects
            # pandas reads in strings as objects by default: https://pandas.pydata.org/docs/user_guide/text.html
            all_elements_are_strings = all([isinstance(element, str) for element in candidate_dataset[:].flat])
            if all_elements_are_strings:
                dtype = np.array([element for element in candidate_dataset[:].flat]).dtype
                chunk_shape = SliceableDataChunkIterator.estimate_default_chunk_shape(
                    chunk_mb=10.0, maxshape=full_shape, dtype=dtype
                )
                buffer_shape = SliceableDataChunkIterator.estimate_default_buffer_shape(
                    buffer_gb=0.5, chunk_shape=chunk_shape, maxshape=full_shape, dtype=dtype
                )
                compression_method = "gzip"
            else:
                raise NotImplementedError(
                    f"Unable to create a `DatasetIOConfiguration` for the dataset at '{location_in_file}'"
                    f"for neurodata object '{neurodata_object}' of type '{type(neurodata_object)}'!"
                )
                # TODO: Add support for compound objects with non-string elements
                # chunk_shape = full_shape  # validate_all_shapes fails if chunk_shape or buffer_shape is None
                # buffer_shape = full_shape
                # compression_method = None
                # warnings.warn(
                #     f"Default chunking and compression options for compound objects are not optimized. "
                #     f"Consider manually specifying DatasetIOConfiguration for dataset at '{location_in_file}'."
                # )

        return cls(
            object_id=neurodata_object.object_id,
            object_name=neurodata_object.name,
            location_in_file=location_in_file,
            dataset_name=dataset_name,
            full_shape=full_shape,
            dtype=dtype,
            chunk_shape=chunk_shape,
            buffer_shape=buffer_shape,
            compression_method=compression_method,
        )
